from pathlib import Path

import pytest

import xorq.api as xo
import xorq.vendor.ibis.expr.operations.relations as rel
from xorq.backends.xorq import Backend
from xorq.caching import (
    ParquetCache,
)


@pytest.fixture
def cached_two(con, pg, tmp_path):
    parquet_cache = ParquetCache.from_kwargs(source=con, relative_path=tmp_path)
    return (
        pg.table("batting")[lambda t: t.yearID > 2014]
        .cache()[lambda t: t.stint == 1]
        .cache(cache=parquet_cache)
    )


@pytest.fixture
def duck_batting_raw(batting_df):
    return xo.duckdb.connect().create_table(
        "batting_df",
        batting_df,
    )


@pytest.fixture
def duck_batting(duck_batting_raw):
    return duck_batting_raw


@pytest.fixture
def cached_two_joined(cached_two, duck_batting):
    return cached_two.join(
        duck_batting,
        duck_batting.columns,
    )


def test_ls_exists(batting):
    attr = getattr(batting, "ls", None)
    assert attr is not None


@pytest.mark.postgres
def test_ls_cache_nodes(cached_two):
    assert len(cached_two.ls.cached_nodes) == 2


@pytest.mark.postgres
def test_cache(cached_two):
    assert cached_two.ls.is_cached
    assert cached_two.ls.cache


@pytest.mark.postgres
def test_caches(cached_two):
    assert cached_two.ls.is_cached
    assert len(cached_two.ls.caches) == len(cached_two.ls.cached_nodes)


@pytest.mark.postgres
def test_backends(duck_batting_raw, cached_two, cached_two_joined):
    assert len(duck_batting_raw.ls.backends) == 1
    assert len(cached_two.ls.backends) == 2
    assert len(cached_two_joined.ls.backends) == 3


@pytest.mark.postgres
def test_is_multiengine(duck_batting_raw, cached_two, cached_two_joined):
    assert not duck_batting_raw.ls.is_multiengine
    assert cached_two.ls.is_multiengine
    assert cached_two_joined.ls.is_multiengine


@pytest.mark.postgres
def test_dts(cached_two, cached_two_joined):
    dts = cached_two.ls.dts
    assert len(dts) == 1
    assert not any(dt.source.name == Backend.name for dt in dts)

    dts = cached_two_joined.ls.dts
    assert len(dts) == 2
    assert not any(dt.source.name == Backend.name for dt in dts)


@pytest.mark.postgres
def test_is_cached(cached_two, cached_two_joined):
    assert cached_two.ls.is_cached
    assert not cached_two_joined.ls.is_cached


@pytest.mark.postgres
def test_has_cached(cached_two, cached_two_joined):
    els = cached_two.ls
    assert els.is_cached and els.has_cached
    els = cached_two_joined.ls
    assert not els.is_cached and els.has_cached


@pytest.mark.postgres
def test_uncached(cached_two):
    assert cached_two.ls.has_cached and not cached_two.ls.uncached.ls.has_cached


@pytest.mark.postgres
def test_uncached_one(cached_two):
    assert (
        cached_two.ls.is_cached
        and not cached_two.ls.uncached_one.ls.is_cached
        and cached_two.ls.uncached_one.ls.has_cached
    )


@pytest.mark.postgres
def test_exists(cached_two):
    cache = cached_two.ls.cache

    assert not cached_two.ls.exists()
    assert not tuple(cache.storage.path.iterdir())

    xo.execute(cached_two)
    assert cached_two.ls.exists()
    assert len(tuple(cache.storage.path.iterdir())) == 1

    (path,) = cache.storage.path.iterdir()
    path.unlink()
    assert not cached_two.ls.exists()


def test_cache_properties(parquet_dir, tmp_path):
    con = xo.connect()
    cache = ParquetCache.from_kwargs(
        source=con, relative_path="./tmp-cache", base_path=tmp_path
    )
    t = xo.deferred_read_parquet(
        parquet_dir.joinpath("batting.parquet"),
        con=con,
        table_name="t",
    )
    expr = (
        t.cache(cache)
        .mutate(a=1)
        .cache(cache)
        .mutate(b=2)
        .cache(cache)
        .mutate(b=3)
        .cache()
    )
    assert not any(p.exists() for p in expr.ls.get_cache_paths())
    expr.execute()
    assert all(p.exists() for p in expr.ls.get_cache_paths())
    (ssn, psn0, psn1, psn2) = expr.ls.cached_nodes
    (ss, ps, *rest) = expr.ls.caches
    assert (ps,) == tuple(set(rest))
    p = cache.storage.path

    # downstream is still cached even if upstream caches don't exist
    (*to_remove, last) = sorted(
        cache.storage.path.iterdir(), key=lambda p: p.stat().st_ctime
    )
    for p in to_remove:
        p.unlink()
    assert all(
        (
            psn0.to_expr().ls.exists(),
            not psn1.to_expr().ls.exists(),
            not psn2.to_expr().ls.exists(),
        )
    )

    # cached_dt differs based on type of Storage
    assert isinstance(psn0.to_expr().ls.cached_dt, xo.expr.relations.Read)
    assert isinstance(ssn.to_expr().ls.cached_dt, rel.DatabaseTable)
    assert all(
        isinstance(psn.to_expr().ls.cache_path, Path) for psn in (psn0, psn1, psn2)
    )
    assert psn0.to_expr().ls.cache_path.exists()
    assert not psn1.to_expr().ls.cache_path.exists()
    assert not psn2.to_expr().ls.cache_path.exists()
    assert ssn.to_expr().ls.cache_path is None



from contextlib import contextmanager

import pytest
import sqlglot as sg
import sqlglot.expressions as sge

import xorq.api as xo
from xorq.vendor.ibis.util import gen_name


SU = pytest.importorskip("xorq.common.utils.snowflake_utils")


@pytest.fixture(scope="session")
def sf_con():
    return xo.snowflake.connect_env_keypair(
        # a database/schema we can trust exists
        database="SNOWFLAKE_SAMPLE_DATA",
        schema="TPCH_SF1",
        create_object_udfs=False,
    )


@pytest.fixture
def temp_catalog(sf_con):
    cat = gen_name("tmp_catalog")

    sf_con.create_catalog(cat)
    assert cat in sf_con.list_catalogs()

    yield cat

    sf_con.drop_catalog(cat)
    assert cat not in sf_con.list_catalogs()


@pytest.fixture
def temp_db(sf_con, temp_catalog):
    database = gen_name("tmp_database")

    sf_con.create_database(database, catalog=temp_catalog)
    assert database in sf_con.list_databases(catalog=temp_catalog)

    yield database

    sf_con.drop_database(database, catalog=temp_catalog)
    assert database not in sf_con.list_databases(catalog=temp_catalog)


@pytest.fixture
def temp_catalog2(sf_con):
    cat = gen_name("tmp_catalog")

    sf_con.create_catalog(cat)
    assert cat in sf_con.list_catalogs()

    yield cat

    sf_con.drop_catalog(cat)
    assert cat not in sf_con.list_catalogs()


@pytest.fixture
def temp_db2(sf_con, temp_catalog2):
    database = gen_name("tmp_database")

    sf_con.create_database(database, catalog=temp_catalog2)
    assert database in sf_con.list_databases(catalog=temp_catalog2)

    yield database

    sf_con.drop_database(database, catalog=temp_catalog2)
    assert database not in sf_con.list_databases(catalog=temp_catalog2)


@contextmanager
def inside_temp_schema(con, temp_catalog, temp_db):
    (prev_catalog, prev_db) = (con.current_catalog, con.current_database)
    con.raw_sql(
        sge.Use(
            kind="SCHEMA", this=sg.table(temp_db, db=temp_catalog, quoted=True)
        ).sql(dialect=con.name),
    )
    try:
        yield
    finally:
        con.raw_sql(
            sge.Use(
                kind="SCHEMA", this=sg.table(prev_db, db=prev_catalog, quoted=True)
            ).sql(dialect=con.name),
        )


def generate_mock_get_conn(original_get_conn, temp_catalog, temp_db):
    @contextmanager
    def mock_get_conn(self, **kwargs):
        kwargs["database"] = "SNOWFLAKE_SAMPLE_DATA"
        kwargs["schema"] = "TPCH_SF1"
        with original_get_conn(self, **kwargs) as conn:
            yield MockSnowFlakeADBCCon(conn, temp_catalog, temp_db)

    return mock_get_conn


class MockSnowFlakeADBCCon:
    def __init__(self, snowflake_adbc_con, temp_catalog, temp_db):
        self.snowflake_adbc_con = snowflake_adbc_con
        self.temp_catalog = temp_catalog
        self.temp_db = temp_db

    @contextmanager
    def cursor(self):
        with self.snowflake_adbc_con.cursor() as cur:
            cur.execute(f'USE SCHEMA "{self.temp_catalog}"."{self.temp_db}"')
            yield cur

    def commit(self):
        self.snowflake_adbc_con.commit()



import functools
from abc import (
    ABC,
    abstractmethod,
)
from typing import Callable

import dask
import pyarrow as pa
import toolz

import xorq.expr.relations as rel
import xorq.vendor.ibis.expr.operations as ops
from xorq.backends.xorq import connect as xo_connect
from xorq.common.utils import classproperty
from xorq.common.utils.func_utils import (
    return_constant,
)
from xorq.common.utils.graph_utils import (
    replace_nodes,
    walk_nodes,
)
from xorq.common.utils.rbr_utils import (
    copy_rbr_batches,
    excepts_print_exc,
    make_filtered_reader,
)
from xorq.vendor import ibis


def schemas_equal(s0, s1):
    def schema_to_dct(s):
        return {name: s.field(name) for name in s.names}

    return schema_to_dct(s0) == schema_to_dct(s1)


def replace_one_unbound(unbound_expr, table):
    # FIXME: consolidate UnboundExprExchanger.set_one_unbound_name and this logic
    (unbound, *rest) = walk_nodes(ops.UnboundTable, unbound_expr)
    if rest:
        raise ValueError(
            "unbound_expr must contain exactly one UnboundTable, but found multiple"
        )
    dt = table.op()
    if not isinstance(dt, (ops.DatabaseTable, rel.Read)):
        raise ValueError(f"table must be a DatabaseTable or Read, got {type(dt)}")
    if not unbound.schema == dt.schema:
        raise ValueError(
            f"unbound schema {unbound.schema} does not match table schema {dt.schema}"
        )

    def _replace_unbound(node, kwargs):
        if isinstance(node, ops.UnboundTable):
            return dt
        elif kwargs:
            return node.__recreate__(kwargs)
        else:
            return node

    return replace_nodes(_replace_unbound, unbound_expr.op()).to_expr()


@excepts_print_exc
def streaming_exchange(
    f, context, reader, writer, options=None, out_schema=None, **kwargs
):
    started = False
    for chunk in (chunk for chunk in reader if chunk.data):
        out = f(chunk.data, metadata=chunk.app_metadata)
        if out_schema is not None:
            assert out_schema == out.schema, (out_schema, "\n", out.schema)

        if not started:
            writer.begin(out.schema, options=options)
            started = True
        writer.write_batch(out)


def find_unbound_next_con(unbound_expr):
    from xorq.common.utils.graph_utils import find_all_sources
    from xorq.common.utils.node_utils import gen_downstream

    match walk_nodes(ops.UnboundTable, unbound_expr):
        case (unbound_op,):
            pass
        case _:
            raise ValueError("unbound_expr must contain exactly one UnboundTable")
    match next(
        filter(None, map(find_all_sources, gen_downstream(unbound_expr, unbound_op))),
        None,
    ):
        case None:
            return None
        case (con, *_) if con.name != "duckdb":
            return con
        case (con, *_) if con.name == "duckdb":
            # duckdb silently hangs when we try to do the replacement
            return None
        case _:
            raise ValueError("unexpected match case in find_unbound_next_con")


@excepts_print_exc
def streaming_expr_exchange(
    unbound_expr, make_connection, context, reader, writer, options=None, **kwargs
):
    filtered_reader = copy_rbr_batches(make_filtered_reader(reader))
    con = find_unbound_next_con(unbound_expr) or make_connection()
    bound_expr = replace_one_unbound(
        unbound_expr,
        con.read_record_batches(filtered_reader),
    )
    started = False
    for batch in bound_expr.to_pyarrow_batches():
        if not started:
            writer.begin(batch.schema, options=options)
            started = True
        writer.write_batch(batch)


class AbstractExchanger(ABC):
    @classproperty
    @abstractmethod
    def exchange_f(cls):
        # return a function with the signature (context, reader, writer, **kwargs)
        def f(context, reader, writer, **kwargs):
            raise NotImplementedError

        return f

    @classproperty
    @abstractmethod
    def schema_in_required(cls):
        pass

    @classproperty
    @abstractmethod
    def schema_in_condition(cls):
        pass

    @classproperty
    @abstractmethod
    def calc_schema_out(cls):
        pass

    @classproperty
    @abstractmethod
    def description(cls):
        pass

    @classproperty
    @abstractmethod
    def command(cls):
        pass

    @classproperty
    def query_result(cls):
        return {
            "schema-in-required": cls.schema_in_required,
            "schema-in-condition": cls.schema_in_condition,
            "calc-schema-out": cls.calc_schema_out,
            "description": cls.description,
            "command": cls.command,
        }


class EchoExchanger(AbstractExchanger):
    @classproperty
    def exchange_f(cls):
        def exchange_echo(context, reader, writer, options=None, **kwargs):
            """Run a simple echo server."""
            started = False
            for chunk in reader:
                if not started and chunk.data:
                    writer.begin(chunk.data.schema, options=options)
                    started = True
                if chunk.app_metadata and chunk.data:
                    writer.write_with_metadata(chunk.data, chunk.app_metadata)
                elif chunk.app_metadata:
                    writer.write_metadata(chunk.app_metadata)
                elif chunk.data:
                    writer.write_batch(chunk.data)
                else:
                    raise AssertionError("Should not happen")

        return exchange_echo

    @classproperty
    def schema_in_required(cls):
        return None

    @classproperty
    def schema_in_condition(cls):
        def condition(schema_in):
            return True

        return condition

    @classproperty
    def calc_schema_out(cls):
        def f(schema_in):
            return schema_in

        return f

    @classproperty
    def description(cls):
        return "echo's data back"

    @classproperty
    def command(cls):
        return "echo"


class PandasUDFExchanger(AbstractExchanger):
    def __init__(self, f, schema_in, name, typ, append=True):
        self.f = f
        self.schema_in = schema_in
        self.name = name
        self.typ = typ
        self.append = append

    @property
    def exchange_f(self):
        def f(batch, metadata=None, **kwargs):
            df = batch.to_pandas()
            series = self.f(df).rename(self.name)
            if self.append:
                out = df.assign(**{series.name: series})
            else:
                out = series.to_frame()
            return pa.RecordBatch.from_pandas(out, preserve_index=False)

        return functools.partial(streaming_exchange, f)

    @property
    def schema_in_required(self):
        return self.schema_in

    @property
    def schema_in_condition(self):
        def condition(schema_in):
            return all(el in schema_in for el in self.schema_in_required)

        return condition

    @property
    def calc_schema_out(self):
        def f(schema_in):
            # FIXME: what to send if schema_in does not match schema_in_required?
            field_schema = ibis.schema({self.name: self.typ})
            if self.append:
                schema_out = schema_in | field_schema
            else:
                schema_out = field_schema
            return schema_out

        return f

    @property
    def description(self):
        return f"a custom udf for {self.f.__name__}"

    @property
    def command(self):
        return f"custom-udf-{self.f.__name__}"

    @property
    def query_result(self):
        return {
            "schema-in-required": self.schema_in_required,
            "schema-in-condition": self.schema_in_condition,
            "calc-schema-out": self.calc_schema_out,
            "description": self.description,
            "command": self.command,
        }


class UnboundExprExchanger(AbstractExchanger):
    def __init__(self, unbound_expr, make_connection=xo_connect):
        self.get_one_unbound(unbound_expr)
        self.unbound_expr = self.set_one_unbound_name(unbound_expr)
        self.make_connection = make_connection
        self._schema_in_required = self.get_one_unbound(self.unbound_expr).schema
        self._schema_in_condition = toolz.curried.operator.eq(self._schema_in_required)
        self._schema_out = self.unbound_expr.schema()

    @staticmethod
    def get_one_unbound(expr):
        (unbound, *rest) = walk_nodes(ops.UnboundTable, expr)
        if rest:
            raise ValueError("expr must only have one unbound table")
        return unbound

    @staticmethod
    def set_one_unbound_name(expr, name="fixed-name"):
        def set_name(op, kwargs):
            if isinstance(op, ops.UnboundTable):
                op = op.copy(name=name)
            if kwargs:
                op = op.__recreate__(kwargs)
            return op

        return replace_nodes(set_name, expr).to_expr()

    @property
    def exchange_f(self):
        return functools.partial(
            streaming_expr_exchange, self.unbound_expr, self.make_connection
        )

    @property
    def schema_in_required(self):
        return self._schema_in_required

    def schema_in_condition(self, schema_in):
        return self._schema_in_condition(schema_in)

    def calc_schema_out(self, schema_in):
        return self._schema_out

    @classproperty
    def description(cls):
        return "run the given unbound expr on the rbr"

    @property
    def command(self):
        return f"execute-unbound-expr-{self.unbound_expr.ls.tokenized}"

    @property
    def query_result(self):
        return {
            "schema-in-required": self._schema_in_required,
            "schema-in-condition": self._schema_in_condition,
            "calc-schema-out": return_constant(self._schema_out),
            "description": self.description,
            "command": self.command,
        }


def make_udxf(
    process_df,
    maybe_schema_in,
    maybe_schema_out,
    name=None,
    description=None,
    command=None,
    do_wraps=True,
):
    def process_batch(process_df, batch, metadata=None, **kwargs):
        df = batch.to_pandas()
        out = process_df(df)
        return pa.RecordBatch.from_pandas(out, preserve_index=False)

    if isinstance(maybe_schema_in, pa.Schema):
        raise ValueError(
            "maybe_schema_in must be an ibis.Schema or Callable, not a pyarrow Schema"
        )
    elif isinstance(maybe_schema_in, ibis.Schema):
        schema_in_required = maybe_schema_in
        schema_in_condition = toolz.curried.operator.eq(schema_in_required)
    elif isinstance(maybe_schema_in, Callable):
        schema_in_required = None
        schema_in_condition = maybe_schema_in
    else:
        raise ValueError(
            f"maybe_schema_in must be an ibis.Schema or Callable, got {type(maybe_schema_in)}"
        )

    out_schema = None
    if isinstance(maybe_schema_out, pa.Schema):
        raise ValueError(
            "maybe_schema_out must be an ibis.Schema or Callable, not a pyarrow Schema"
        )
    elif isinstance(maybe_schema_out, ibis.Schema):
        calc_schema_out = return_constant(maybe_schema_out)
        out_schema = maybe_schema_out
    elif isinstance(maybe_schema_out, Callable):
        calc_schema_out = maybe_schema_out
    else:
        raise ValueError(
            f"maybe_schema_out must be an ibis.Schema or Callable, got {type(maybe_schema_out)}"
        )

    if do_wraps:
        exchange_f = excepts_print_exc(
            functools.partial(
                streaming_exchange,
                functools.partial(process_batch, process_df),
                out_schema=out_schema.to_pyarrow() if out_schema else None,
            ),
            Exception,
        )
    else:
        exchange_f = process_df

    name = name or process_df.__name__
    typ = type(
        name,
        (AbstractExchanger,),
        {
            "exchange_f": exchange_f,
            "schema_in_required": schema_in_required,
            "schema_in_condition": schema_in_condition,
            "calc_schema_out": calc_schema_out,
            "description": description or name,
            "command": command or dask.base.tokenize(process_df),
        },
    )
    return typ


exchangers = {exchanger.command: exchanger for exchanger in (EchoExchanger,)}



import base64
import logging
import secrets
import threading

import pyarrow as pa
import pyarrow.flight
from cloudpickle import loads

import xorq.flight.action as A
import xorq.flight.exchanger as E
from xorq.common.utils.func_utils import (
    maybe_log_excepts,
)
from xorq.common.utils.rbr_utils import (
    copy_rbr_batches,
)
from xorq.flight.metrics import (
    InstrumentedReader,
    InstrumentedWriter,
    instrument_reader,
    instrument_rpc,
)


logger = logging.getLogger(__name__)


class BasicAuthServerMiddlewareFactory(pa.flight.ServerMiddlewareFactory):
    """
    Middleware that implements username-password authentication.

    Parameters
    ----------
    creds: Dict[str, str]
        A dictionary of username-password values to accept.
    """

    def __init__(self, creds):
        self.creds = creds
        # Map generated bearer tokens to users
        self.tokens = {}

    def start_call(self, info, headers):
        """Validate credentials at the start of every call."""
        # Search for the authentication header (case-insensitive)
        auth_header = None
        for header in headers:
            if header.lower() == "authorization":
                auth_header = headers[header][0]
                break

        if not auth_header:
            raise pa.flight.FlightUnauthenticatedError("No credentials supplied")

        # The header has the structure "AuthType TokenValue", e.g.
        # "Basic <encoded username+password>" or "Bearer <random token>".
        auth_type, _, value = auth_header.partition(" ")

        if auth_type == "Basic":
            # Initial "login". The user provided a username/password
            # combination encoded in the same way as HTTP Basic Auth.
            decoded = base64.b64decode(value).decode("utf-8")
            username, _, password = decoded.partition(":")
            if not password or password != self.creds.get(username):
                raise pa.flight.FlightUnauthenticatedError(
                    "Unknown user or invalid password"
                )
            # Generate a secret, random bearer token for future calls.
            token = secrets.token_urlsafe(32)
            self.tokens[token] = username
            return BasicAuthServerMiddleware(token)
        elif auth_type == "Bearer":
            # An actual call. Validate the bearer token.
            username = self.tokens.get(value)
            if username is None:
                raise pa.flight.FlightUnauthenticatedError("Invalid token")
            return BasicAuthServerMiddleware(value)

        raise pa.flight.FlightUnauthenticatedError("No credentials supplied")


class BasicAuthServerMiddleware(pa.flight.ServerMiddleware):
    """Middleware that implements username-password authentication."""

    def __init__(self, token):
        self.token = token

    def sending_headers(self):
        """Return the authentication token to the client."""
        return {"authorization": f"Bearer {self.token}"}


class NoOpAuthHandler(pa.flight.ServerAuthHandler):
    """
    A handler that implements username-password authentication.

    This is required only so that the server will respond to the internal
    Handshake RPC call, which the client calls when authenticate_basic_token
    is called. Otherwise, it should be a no-op as the actual authentication is
    implemented in middleware.
    """

    def authenticate(self, outgoing, incoming):
        pass

    def is_valid(self, token):
        return ""


class FlightServerDelegate(pyarrow.flight.FlightServerBase):
    def __init__(
        self,
        con_callable,
        location=None,
        tls_certificates=None,
        verify_client=False,
        root_certificates=None,
        auth_handler=None,
        middleware=None,
    ):
        super(FlightServerDelegate, self).__init__(
            location=location,
            auth_handler=auth_handler,
            tls_certificates=tls_certificates,
            verify_client=verify_client,
            root_certificates=root_certificates,
            middleware=middleware,
        )
        self._conn = con_callable()
        self._location = location
        self.exchangers = dict(E.exchangers)
        self.actions = A.actions
        self.lock = threading.Lock()

    def _make_flight_info(self, query):
        """
        Create Flight info for a given SQL query

        Args:
            query: SQL query string
        """
        # Execute query to get schema and metadata
        kwargs = loads(query)
        expr = kwargs.pop("expr")
        schema = expr.as_table().schema().to_pyarrow()
        descriptor = pyarrow.flight.FlightDescriptor.for_command(query)
        endpoints = [pyarrow.flight.FlightEndpoint(query, [self._location])]

        return pyarrow.flight.FlightInfo(schema, descriptor, endpoints, -1, -1)

    def get_flight_info(self, context, descriptor):
        """
        Get info about a specific query
        """
        query = descriptor.command
        return self._make_flight_info(query)

    @instrument_rpc("do_get")
    @maybe_log_excepts
    def do_get(self, context, rec, ticket):
        kwargs = loads(ticket.ticket)
        expr = kwargs.pop("expr")
        with self.lock:
            rbr = self._conn.to_pyarrow_batches(expr)
        rbr = instrument_reader(rbr, rec, direction="out")
        return pyarrow.flight.RecordBatchStream(rbr)

    @instrument_rpc("do_put")
    @maybe_log_excepts
    def do_put(self, context, rec, descriptor, reader, writer):
        reader = instrument_reader(reader, rec, direction="in")

        cmd = loads(descriptor.command)
        if isinstance(cmd, bytes):
            table = cmd.decode("utf-8")
            kwargs = {}
        else:
            table = cmd["table_name"]
            kwargs = cmd.get("kwargs", {})

        data = copy_rbr_batches(reader).read_all()
        with self.lock:
            if table in self._conn.tables:
                self._conn.insert(table, data, **kwargs)
            else:
                self._conn.create_table(table, data, **kwargs)

    @maybe_log_excepts
    def do_action(self, context, action):
        with self.lock:
            cls = self.actions.get(action.type)
            if cls:
                logger.info(f"doing action: {action.type}")
                yield from cls.do_action(self, context, action)
            else:
                raise KeyError("Unknown action {!r}".format(action.type))

    @instrument_rpc("do_exchange")
    @maybe_log_excepts
    def do_exchange(self, context, rec, descriptor, reader, writer):
        with self.lock:
            cmd = descriptor.command.decode()
            handler = self.exchangers.get(cmd)
            if handler is None:
                raise pa.ArrowInvalid(f"Unknown exchange: {cmd}")

            reader = InstrumentedReader(reader, rec, direction="in")
            iw = InstrumentedWriter(writer, rec, direction="out")

            return handler.exchange_f(context, reader, iw)



from __future__ import annotations

import decimal
import math
import operator
from operator import and_, lshift, or_, rshift, xor
from typing import Callable

import numpy as np
import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
from xorq.tests.util import assert_series_equal, default_series_rename
from xorq.vendor import ibis
from xorq.vendor.ibis import _
from xorq.vendor.ibis import literal as L
from xorq.vendor.ibis.expr import datatypes as dt


@pytest.mark.parametrize(
    ("expr",),
    [
        param(
            xo.literal(1, type=dt.int8),
            id="int8",
        ),
        param(
            xo.literal(1, type=dt.int16),
            id="int16",
        ),
        param(
            xo.literal(1, type=dt.int32),
            id="int32",
        ),
        param(
            xo.literal(1, type=dt.int64),
            id="int64",
        ),
        param(
            xo.literal(1, type=dt.uint8),
            id="uint8",
        ),
        param(
            xo.literal(1, type=dt.uint16),
            id="uint16",
        ),
        param(
            xo.literal(1, type=dt.uint32),
            id="uint32",
        ),
        param(
            xo.literal(1, type=dt.uint64),
            id="uint64",
        ),
        param(
            xo.literal(1, type=dt.float32),
            id="float32",
        ),
        param(
            xo.literal(1, type=dt.float64),
            id="float64",
        ),
    ],
)
def test_numeric_literal(con, expr):
    result = con.execute(expr)
    assert result == 1


@pytest.mark.parametrize(
    ("expr", "expected_result"),
    [
        param(
            xo.literal(decimal.Decimal("1.1"), type=dt.decimal),
            decimal.Decimal("1.1"),
            id="default",
        ),
        param(
            xo.literal(decimal.Decimal("1.1"), type=dt.Decimal(38, 9)),
            decimal.Decimal("1.1"),
            id="decimal-small",
        ),
    ],
)
def test_decimal_literal(con, expr, expected_result):
    result = con.execute(expr)
    if type(expected_result) in (float, decimal.Decimal) and math.isnan(
        expected_result
    ):
        assert math.isnan(result) and type(result) is type(expected_result)
    else:
        assert result == expected_result


@pytest.mark.parametrize(
    ("operand_fn", "expected_operand_fn"),
    [
        param(
            lambda t: t.float_col,
            lambda t: t.float_col,
            id="float-column",
        ),
        param(
            lambda t: t.double_col,
            lambda t: t.double_col,
            id="double-column",
        ),
        param(
            lambda t: xo.literal(1.3),
            lambda t: 1.3,
            id="float-literal",
        ),
        param(
            lambda t: xo.literal(np.nan),
            lambda t: np.nan,
            id="nan-literal",
        ),
        param(
            lambda t: xo.literal(np.inf),
            lambda t: np.inf,
            id="inf-literal",
        ),
        param(
            lambda t: xo.literal(-np.inf),
            lambda t: -np.inf,
            id="-inf-literal",
        ),
    ],
)
@pytest.mark.parametrize(
    ("expr_fn", "expected_expr_fn"),
    [
        param(
            operator.methodcaller("isnan"),
            np.isnan,
            id="isnan",
        ),
    ],
)
def test_isnan_isinf(
    con,
    alltypes,
    alltypes_df,
    operand_fn,
    expected_operand_fn,
    expr_fn,
    expected_expr_fn,
):
    expr = expr_fn(operand_fn(alltypes)).name("tmp")
    expected = expected_expr_fn(expected_operand_fn(alltypes_df))

    result = con.execute(expr)

    if isinstance(expected, pd.Series):
        assert_series_equal(result, expected)
    else:
        try:
            assert result == expected
        except ValueError:
            assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(L(-5).abs(), 5, id="abs-neg"),
        param(L(5).abs(), 5, id="abs"),
        param(
            L(5.5).round(),
            6.0,
            id="round",
        ),
        param(
            L(5.556).round(2),
            5.56,
            id="round-digits",
        ),
        param(L(5.556).ceil(), 6.0, id="ceil"),
        param(L(5.556).floor(), 5.0, id="floor"),
        param(
            L(5.556).exp(),
            math.exp(5.556),
            id="exp",
        ),
        param(
            L(5.556).sign(),
            1,
            id="sign-pos",
        ),
        param(
            L(-5.556).sign(),
            -1,
            id="sign-neg",
        ),
        param(
            L(0).sign(),
            0,
            id="sign-zero",
        ),
        param(L(5.556).sqrt(), math.sqrt(5.556), id="sqrt"),
        param(
            L(5.556).log(2),
            math.log2(5.556),
            id="log-base",
        ),
        param(
            L(5.556).ln(),
            math.log(5.556),
            id="ln",
        ),
        param(
            L(5.556).log2(),
            math.log2(5.556),
            id="log2",
        ),
        param(
            L(5.556).log10(),
            math.log10(5.556),
            id="log10",
        ),
        param(
            L(5.556).radians(),
            math.radians(5.556),
            id="radians",
        ),
        param(
            L(5.556).degrees(),
            math.degrees(5.556),
            id="degrees",
        ),
        param(
            L(11) % 3,
            11 % 3,
            id="mod",
        ),
        param(
            xo.greatest(L(10), L(1)),
            10,
            id="greatest",
        ),
        param(
            xo.least(L(10), L(1)),
            1,
            id="least",
        ),
    ],
)
def test_math_functions_literals(con, expr, expected):
    result = con.execute(expr.name("tmp"))
    if isinstance(result, decimal.Decimal):
        assert result == decimal.Decimal(str(expected))
    else:
        np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(L(0.0).acos(), math.acos(0.0), id="acos"),
        param(L(0.0).asin(), math.asin(0.0), id="asin"),
        param(L(0.0).atan(), math.atan(0.0), id="atan"),
        param(L(0.0).atan2(1.0), math.atan2(0.0, 1.0), id="atan2"),
        param(L(0.0).cos(), math.cos(0.0), id="cos"),
        param(L(1.0).cot(), 1.0 / math.tan(1.0), id="cot"),
        param(L(0.0).sin(), math.sin(0.0), id="sin"),
        param(L(0.0).tan(), math.tan(0.0), id="tan"),
    ],
)
def test_trig_functions_literals(con, expr, expected):
    result = con.execute(expr.name("tmp"))
    assert pytest.approx(result) == expected


@pytest.mark.parametrize(
    ("expr", "expected_fn"),
    [
        param(_.dc.acos(), np.arccos, id="acos"),
        param(_.dc.asin(), np.arcsin, id="asin"),
        param(_.dc.atan(), np.arctan, id="atan"),
        param(_.dc.atan2(_.dc), lambda c: np.arctan2(c, c), id="atan2"),
        param(_.dc.cos(), np.cos, id="cos"),
        param(_.dc.cot(), lambda c: 1.0 / np.tan(c), id="cot"),
        param(_.dc.sin(), np.sin, id="sin"),
        param(_.dc.tan(), np.tan, id="tan"),
    ],
)
def test_trig_functions_columns(expr, alltypes, alltypes_df, expected_fn):
    dc_max = alltypes_df.double_col.max()
    expr = alltypes.mutate(dc=(_.double_col / dc_max).nullif(0)).select(tmp=expr)
    result = expr.tmp.to_pandas()
    expected = expected_fn(
        (alltypes_df.double_col / dc_max).replace(0.0, np.nan)
    ).rename("tmp")
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("expr_fn", "expected_fn"),
    [
        param(
            lambda t: (-t.double_col).abs(),
            lambda t: (-t.double_col).abs(),
            id="abs-neg",
        ),
        param(
            lambda t: t.double_col.abs(),
            lambda t: t.double_col.abs(),
            id="abs",
        ),
        param(
            lambda t: t.double_col.ceil(),
            lambda t: np.ceil(t.double_col).astype("int64"),
            id="ceil",
        ),
        param(
            lambda t: t.double_col.floor(),
            lambda t: np.floor(t.double_col).astype("int64"),
            id="floor",
        ),
        param(
            lambda t: t.double_col.sign(),
            lambda t: np.sign(t.double_col),
            id="sign",
        ),
        param(
            lambda t: (-t.double_col).sign(),
            lambda t: np.sign(-t.double_col),
            id="sign-negative",
        ),
    ],
)
def test_simple_math_functions_columns(
    con, alltypes, alltypes_df, expr_fn, expected_fn
):
    expr = expr_fn(alltypes).name("tmp")
    expected = expected_fn(alltypes_df)
    result = con.execute(expr)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("expr_fn", "expected_fn"),
    [
        param(
            lambda t: t.double_col.add(1).sqrt(),
            lambda t: np.sqrt(t.double_col + 1),
            id="sqrt",
        ),
        param(
            lambda t: t.double_col.add(1).exp(),
            lambda t: np.exp(t.double_col + 1),
            id="exp",
        ),
        param(
            lambda t: t.double_col.add(1).log(2),
            lambda t: np.log2(t.double_col + 1),
            id="log2",
        ),
        param(
            lambda t: t.double_col.add(1).ln(),
            lambda t: np.log(t.double_col + 1),
            id="ln",
        ),
        param(
            lambda t: t.double_col.add(1).log10(),
            lambda t: np.log10(t.double_col + 1),
            id="log10",
        ),
    ],
)
def test_complex_math_functions_columns(
    con, alltypes, alltypes_df, expr_fn, expected_fn
):
    expr = expr_fn(alltypes).name("tmp")
    expected = expected_fn(alltypes_df)
    result = con.execute(expr)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("expr_fn", "expected_fn"),
    [
        param(
            lambda t: xo.least(t.bigint_col, t.int_col),
            lambda t: pd.Series(list(map(min, t.bigint_col, t.int_col))),
            id="least-all-columns",
        ),
        param(
            lambda t: xo.least(t.bigint_col, t.int_col, -2),
            lambda t: pd.Series(list(map(min, t.bigint_col, t.int_col, [-2] * len(t)))),
            id="least-scalar",
        ),
        param(
            lambda t: xo.greatest(t.bigint_col, t.int_col),
            lambda t: pd.Series(list(map(max, t.bigint_col, t.int_col))),
            id="greatest-all-columns",
        ),
        param(
            lambda t: xo.greatest(t.bigint_col, t.int_col, -2),
            lambda t: pd.Series(list(map(max, t.bigint_col, t.int_col, [-2] * len(t)))),
            id="greatest-scalar",
        ),
    ],
)
def test_backend_specific_numerics(
    con, alltypes_df, alltypes, expr_fn: Callable, expected_fn: Callable
):
    expr = expr_fn(alltypes)
    result = default_series_rename(con.execute(expr.name("tmp")))
    expected = default_series_rename(expected_fn(alltypes_df))
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    "op",
    [
        operator.add,
        operator.sub,
        operator.mul,
        operator.truediv,
        operator.floordiv,
        operator.pow,
    ],
    ids=lambda op: op.__name__,
)
def test_binary_arithmetic_operations(alltypes, alltypes_df, op):
    smallint_col = alltypes.smallint_col + 1  # make it nonzero
    smallint_series = alltypes_df.smallint_col + 1

    expr = op(alltypes.double_col, smallint_col).name("tmp")

    result = expr.execute()
    expected = op(alltypes_df.double_col, smallint_series)
    if op is operator.floordiv:
        # defined in ops.FloorDivide.output_type
        # -> returns int64 whereas pandas float64
        result = result.astype("float64")

    expected = expected.astype("float64")
    assert_series_equal(result, expected, check_exact=False)


def test_mod(alltypes, alltypes_df):
    expr = operator.mod(alltypes.smallint_col, alltypes.smallint_col + 1).name("tmp")

    result = expr.execute()
    expected = operator.mod(alltypes_df.smallint_col, alltypes_df.smallint_col + 1)
    assert_series_equal(result, expected, check_dtype=False)


def test_floating_mod(alltypes, alltypes_df):
    expr = operator.mod(alltypes.double_col, alltypes.smallint_col + 1).name("tmp")

    result = expr.execute()
    expected = operator.mod(alltypes_df.double_col, alltypes_df.smallint_col + 1)
    assert_series_equal(result, expected, check_exact=False)


@pytest.mark.parametrize(
    ("column", "denominator"),
    [
        param(
            "tinyint_col",
            0,
        ),
        param(
            "smallint_col",
            0,
        ),
        param(
            "int_col",
            0,
        ),
        param(
            "bigint_col",
            0,
        ),
        param(
            "float_col",
            0,
        ),
        param(
            "double_col",
            0,
        ),
        param(
            "tinyint_col",
            0.0,
        ),
        param(
            "smallint_col",
            0.0,
        ),
        param(
            "int_col",
            0.0,
        ),
        param(
            "bigint_col",
            0.0,
        ),
        param(
            "float_col",
            0.0,
        ),
        param(
            "double_col",
            0.0,
        ),
    ],
)
def test_divide_by_zero(alltypes, alltypes_df, column, denominator):
    expr = alltypes[column] / denominator
    result = expr.name("tmp").execute()

    expected = alltypes_df[column].div(denominator)
    expected = expected.astype("float64")

    assert_series_equal(result.astype("float64"), expected)


def test_random(con):
    expr = xo.random()
    result = con.execute(expr)
    assert isinstance(result, float)
    assert 0 <= result <= 1


def test_histogram(con, alltypes):
    n = 10
    hist = con.execute(alltypes.int_col.histogram(n).name("hist"))
    vc = hist.value_counts().sort_index()
    vc_np, _bin_edges = np.histogram(alltypes.int_col.execute(), bins=n)
    assert vc.tolist() == vc_np.tolist()


@pytest.mark.parametrize("const", ["pi", "e"])
def test_constants(con, const):
    expr = getattr(ibis, const)
    result = con.execute(expr)
    assert pytest.approx(result) == getattr(math, const)


@pytest.mark.parametrize("op", [and_, or_, xor])
@pytest.mark.parametrize(
    ("left_fn", "right_fn"),
    [
        param(lambda t: t.int_col, lambda t: t.int_col, id="col_col"),
        param(lambda _: 3, lambda t: t.int_col, id="scalar_col"),
        param(lambda t: t.int_col, lambda _: 3, id="col_scalar"),
    ],
)
def test_bitwise_columns(con, alltypes, alltypes_df, op, left_fn, right_fn):
    expr = op(left_fn(alltypes), right_fn(alltypes)).name("tmp")
    result = con.execute(expr)

    expected = op(left_fn(alltypes_df), right_fn(alltypes_df)).rename("tmp")
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("op", "left_fn", "right_fn"),
    [
        param(
            lshift,
            lambda t: t.int_col,
            lambda t: t.int_col,
            id="lshift_col_col",
        ),
        param(
            lshift,
            lambda _: 3,
            lambda t: t.int_col,
            id="lshift_scalar_col",
        ),
        param(lshift, lambda t: t.int_col, lambda _: 3, id="lshift_col_scalar"),
        param(rshift, lambda t: t.int_col, lambda t: t.int_col, id="rshift_col_col"),
        param(rshift, lambda _: 3, lambda t: t.int_col, id="rshift_scalar_col"),
        param(rshift, lambda t: t.int_col, lambda _: 3, id="rshift_col_scalar"),
    ],
)
def test_bitwise_shift(alltypes, alltypes_df, op, left_fn, right_fn):
    expr = op(left_fn(alltypes), right_fn(alltypes)).name("tmp")
    result = expr.execute()

    pandas_left = getattr(left := left_fn(alltypes_df), "values", left)
    pandas_right = getattr(right := right_fn(alltypes_df), "values", right)
    expected = pd.Series(
        op(pandas_left, pandas_right),
        name="tmp",
        dtype="int64",
    )
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    "op",
    [
        and_,
        or_,
        xor,
        lshift,
        rshift,
    ],
)
@pytest.mark.parametrize(
    ("left", "right"),
    [param(4, L(2), id="int_col"), param(L(4), 2, id="col_int")],
)
def test_bitwise_scalars(con, op, left, right):
    expr = op(left, right)
    result = con.execute(expr)
    expected = op(4, 2)
    assert result == expected


def test_bitwise_not_scalar(con):
    expr = ~L(2)
    result = con.execute(expr)
    expected = -3
    assert result == expected


def test_bitwise_not_col(alltypes, alltypes_df):
    expr = (~alltypes.int_col).name("tmp")
    result = expr.execute()
    expected = ~alltypes_df.int_col
    assert_series_equal(result, expected.rename("tmp"))


@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        param(lambda x: x.clip(lower=0), lambda x: x.clip(lower=0), id="lower-int"),
        param(
            lambda x: x.clip(lower=0.0), lambda x: x.clip(lower=0.0), id="lower-float"
        ),
        param(lambda x: x.clip(upper=0), lambda x: x.clip(upper=0), id="upper-int"),
        param(
            lambda x: x.clip(lower=x - 1, upper=x + 1),
            lambda x: x.clip(lower=x - 1, upper=x + 1),
            id="lower-upper-expr",
        ),
        param(
            lambda x: x.clip(lower=0, upper=1),
            lambda x: x.clip(lower=0, upper=1),
            id="lower-upper-int",
        ),
        param(
            lambda x: x.clip(lower=0, upper=1.0),
            lambda x: x.clip(lower=0, upper=1.0),
            id="lower-upper-float",
        ),
        param(
            lambda x: x.nullif(1).clip(lower=0),
            lambda x: x.where(x != 1).clip(lower=0),
            id="null-lower",
        ),
        param(
            lambda x: x.nullif(1).clip(upper=0),
            lambda x: x.where(x != 1).clip(upper=0),
            id="null-upper",
        ),
    ],
)
def test_clip(alltypes, alltypes_df, ibis_func, pandas_func):
    result = ibis_func(alltypes.int_col).execute()
    expected = pandas_func(alltypes_df.int_col).astype(result.dtype)
    assert_series_equal(result, expected, check_names=False)



import functools
import inspect
import operator
import os
import types
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Iterable, List, Tuple, Union

import toolz

import xorq.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor.ibis.common.annotations import Argument
from xorq.vendor.ibis.common.collections import FrozenDict
from xorq.vendor.ibis.common.patterns import pattern, replace
from xorq.vendor.ibis.expr.operations.udf import InputType, ScalarUDF
from xorq.vendor.ibis.expr.rules import ValueOf
from xorq.vendor.ibis.util import Namespace


if TYPE_CHECKING:
    from quickgrove import PyGradientBoostedDecisionTrees

SUPPORTED_TYPES = {
    "float": dt.float64,
    "i": dt.boolean,
    "int": dt.int64,
}

p = Namespace(pattern, module=ops)


def _calculate_bounds(
    test_sizes: Iterable[float] | float,
) -> Tuple[Tuple[float, float]]:
    """
    Calculates the cumulative sum of test_sizes and generates bounds for splitting data.

    Parameters
    ----------
    test_sizes : Iterable[float] | float
        An iterable of floats representing the desired proportions for data splits.
        Each value should be between 0 and 1, and their sum must equal 1. The
        order of test sizes determines the order of the generated subsets. If float is passed
        it assumes that the value is for the test size and that a tradition tain test split of (1-test_size, test_size) is returned.

    Returns
    -------
    Tuple[Tuple[float, float]]
        A Tuple of tuples, where each tuple contains two floats representing the
        lower and upper bounds for a split. These bounds are calculated based on
        the cumulative sum of the `test_sizes`.
    """
    # Convert to traditional train test split
    if isinstance(test_sizes, float):
        test_sizes = (1 - test_sizes, test_sizes)

    test_sizes = tuple(test_sizes)

    if not all(isinstance(test_size, float) for test_size in test_sizes):
        raise ValueError("Test size must be float.")

    if not all((0 < test_size < 1) for test_size in test_sizes):
        raise ValueError("test size should be a float between 0 and 1.")

    try:
        import pandas as pd

        pd._testing.assert_almost_equal(sum(test_sizes), 1)
    except AssertionError as err:
        raise ValueError("Test sizes must sum to 1") from err

    cumulative_sizes = tuple(toolz.accumulate(operator.add, (0,) + test_sizes))
    bounds = tuple(zip(cumulative_sizes[:-1], cumulative_sizes[1:]))
    return bounds


def fields_to_parameters(fields: dict) -> List[inspect.Parameter]:
    """Convert a field dict into `inspect.Parameter` objects for constructing function signatures."""

    parameters = []
    for name, arg in fields.items():
        param = inspect.Parameter(
            name,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=arg.typehint,
        )
        parameters.append(param)
    return parameters


class UDFWrapper:
    """Wrapper to hold the constructed function, model, and path for nice representations."""

    def __init__(
        self,
        func: Callable,
        model: "PyGradientBoostedDecisionTrees",
        model_path: Union[str, Path, None],
        required_features: List[str] = None,
    ):
        self.func = func
        self.model = model
        self.model_path = model_path
        self.required_features = required_features or []
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        path_str = f"\nModel path: {self.model_path}" if self.model_path else ""
        return f"{str(self.model)}{path_str}\nSignature: {str(self.func.__signature__)}"

    def on_expr(self, e: ir.Table, **kwargs) -> ir.Column:
        """
        Call the UDF by extracting the required columns from the table expression `e`.
        E.g., e['col1'], e['col2'], etc.
        """
        columns = [e[col] for col in self.required_features]
        return self(*columns, **kwargs)


def _load_quickgrove_model(
    model_or_path: Union[str, Path, "PyGradientBoostedDecisionTrees"],
) -> Tuple["PyGradientBoostedDecisionTrees", Union[str, Path, None]]:
    """Load the model if a path is provided; otherwise, pass through the model."""

    import quickgrove

    if isinstance(model_or_path, (str, Path)):
        model = quickgrove.json_load(model_or_path)
        return model, model_or_path
    else:
        return model_or_path, None


def _validate_model_features(
    model: "PyGradientBoostedDecisionTrees", supported_types: dict
) -> None:
    """Raise an error if the model has unsupported feature types."""
    schema = [model.feature_names[i] for i in sorted(model.required_features)]
    feature_types = [model.feature_types[i] for i in sorted(model.required_features)]

    unsupported = [
        f"{name}: {type_}"
        for name, type_ in zip(schema, feature_types)
        if type_ not in supported_types
    ]
    if unsupported:
        raise ValueError(f"Unsupported feature types: {', '.join(unsupported)}")


def _create_udf_node(
    model: "PyGradientBoostedDecisionTrees",
    fn_from_arrays: Callable,
    fields: dict,
    udf_name: str,
    extra_meta: dict = None,
) -> type:
    """Dynamically create a subclass of ScalarUDF for a quickgrove-based function."""

    meta = {
        "dtype": dt.float32,
        "__input_type__": InputType.PYARROW,
        "__func__": property(lambda self: fn_from_arrays),
        "__config__": FrozenDict(volatility="immutable"),
        "__udf_namespace__": p,
        "__module__": "xorq.expr.ml.quickgrove_lib",
        "__func_name__": udf_name,
        "__fields__": fields,
        "model": model,
    }
    if extra_meta:
        meta.update(extra_meta)

    node = type(udf_name, (ScalarUDF,), {**fields, **meta})
    return node


def _all_predicates_are_features(filter_op: ops.Filter, model) -> bool:
    feature_set = set(model.feature_names)
    for pred in filter_op.predicates:
        if not (
            isinstance(pred, (ops.Less, ops.Greater, ops.GreaterEqual))
            and isinstance(pred.left, ops.Field)
            and pred.left.name in feature_set
        ):
            return False
    return True


def _extract_model_name(model_path):
    return os.path.splitext(os.path.basename(model_path))[0]


def _create_udf_function(node_cls: type) -> Callable:
    """
    Create a function that, when called, instantiates `node_cls` and returns the Ibis expression.
    Attach an `inspect.Signature` for introspection.
    """
    fn_from_arrays = node_cls.__func__.fget(None)

    @functools.wraps(fn_from_arrays)
    def construct(*args, **kwargs):
        return node_cls(*args, **kwargs).to_expr()

    signature = inspect.Signature(
        fields_to_parameters(node_cls.__fields__),
        return_annotation=dt.float32,
    )

    def actual_func(*args, **kwargs):
        return construct(*args, **kwargs)

    new_func = types.FunctionType(
        actual_func.__code__,
        actual_func.__globals__,
        name=node_cls.__func_name__,
        argdefs=actual_func.__defaults__,
        closure=actual_func.__closure__,
    )
    new_func.__signature__ = signature
    return new_func


def make_quickgrove_udf(
    model_or_path: Union[str, Path, "PyGradientBoostedDecisionTrees"], model_name=None
) -> UDFWrapper:
    """
    Create a UDF from a quickgrove (quickgrove) model.

    :param model_or_path: Either an already-loaded model or a path to the model.
    :param model_name: Alternative model name, if the name from path is not a valid Python identifier.
    :return:
    """

    model, model_path = _load_quickgrove_model(model_or_path)

    if not model_name:
        new_model_name = _extract_model_name(model_path)
        if not new_model_name.isidentifier():
            raise ValueError(
                f"The argument model_name was {model_name} and the name extracted from the path is not a valid Python identifier"
            )
        model_name = new_model_name
    elif not model_name.isidentifier():
        raise ValueError("The argument model_name is not a valid Python identifier")

    _validate_model_features(model, SUPPORTED_TYPES)

    required_features = sorted(model.required_features)
    schema = [model.feature_names[i] for i in required_features]
    feature_types = [model.feature_types[i] for i in required_features]

    fields = {
        name: Argument(
            pattern=ValueOf(SUPPORTED_TYPES[type_]),
            typehint=SUPPORTED_TYPES[type_],
        )
        for name, type_ in zip(schema, feature_types)
    }

    def fn_from_arrays(*arrays):
        return model.predict_arrays(list(arrays))

    udf_func = _create_udf_function(
        _create_udf_node(
            model=model,
            fn_from_arrays=fn_from_arrays,
            fields=fields,
            udf_name=model_name,
            extra_meta={"model_path": model_path},
        )
    )

    return UDFWrapper(
        udf_func,
        model,
        model_path,
        required_features=schema,
    )


def collect_predicates(filter_op: ops.Filter) -> List[dict]:
    """Extract pushdown predicates that compare a base column (ops.Field) to a literal.
    Ignore any predicates referencing a UDF output."""

    parent_op = filter_op.parent

    predicates = []
    for pred in filter_op.predicates:
        if isinstance(pred, (ops.Less, ops.Greater, ops.GreaterEqual)):
            if isinstance(pred.left, ops.Field) and isinstance(pred.right, ops.Literal):
                field_name = pred.left.name
                if isinstance(parent_op.values.get(field_name), ops.ScalarUDF):
                    raise ValueError(
                        f"Unsupported predicate on UDF column: {field_name}"
                    )

                predicates.append(
                    {
                        "column": field_name,
                        "op": pred.__class__.__name__,
                        "value": pred.right.value,
                    }
                )
    return predicates


def make_pruned_udf(
    original_udf: UDFWrapper,
    predicates: List[dict],
) -> UDFWrapper:
    """
    Create a new pruned UDF from an existing UDF and a list of predicates.
    `original_udf` must have a `.model` attribute.
    """

    from quickgrove import Feature

    model = original_udf.model
    pred_feature_names = {pred["column"] for pred in predicates}
    model_feature_names = set(model.feature_names)

    if not pred_feature_names.issubset(model_feature_names):
        warnings.warn(
            "Feature not found in predicates, skipping pruning...",
            UserWarning,
            stacklevel=2,
        )
        return original_udf

    pruned_model = model.prune(
        [
            Feature(pred["column"]) < pred["value"]
            if pred["op"] == "Less"
            else Feature(pred["column"]) >= pred["value"]
            if pred["op"] == "Greater"
            else Feature(pred["column"]) >= pred["value"]
            for pred in predicates
        ]
    )

    _validate_model_features(pruned_model, SUPPORTED_TYPES)

    required_features = sorted(pruned_model.required_features)
    schema = [pruned_model.feature_names[i] for i in required_features]
    feature_types = [pruned_model.feature_types[i] for i in required_features]

    fields = {
        name: Argument(
            pattern=ValueOf(SUPPORTED_TYPES[type_]), typehint=SUPPORTED_TYPES[type_]
        )
        for name, type_ in zip(schema, feature_types)
    }

    def fn_from_arrays(*arrays):
        return pruned_model.predict_arrays(list(arrays))

    udf_func = _create_udf_function(
        _create_udf_node(
            model=pruned_model,
            fn_from_arrays=fn_from_arrays,
            fields=fields,
            udf_name=original_udf.__func_name__ + "_pruned",
        )
    )

    return UDFWrapper(
        udf_func,
        pruned_model,
        original_udf.model_path,
        required_features=schema,
    )


@replace(p.Filter(p.Project))
def prune_quickgrove_model(_, **kwargs):
    """Rewrite rule to prune quickgrove model based on filter predicates."""

    parent_op = _.parent
    predicates = collect_predicates(_)
    if not predicates:
        return _

    new_values = {}
    pruned_udf_wrapper = None

    for name, value in parent_op.values.items():
        if isinstance(value, ops.ScalarUDF) and hasattr(value, "model"):
            # return filter op if predicates are not in filter
            if not _all_predicates_are_features(_, value.model):
                return _

            pruned_udf_wrapper = make_pruned_udf(value, predicates)
            pruned_model = pruned_udf_wrapper.model
            required_features = [
                pruned_model.feature_names[i] for i in pruned_model.required_features
            ]
            required_features = [
                pruned_model.feature_names[i] for i in pruned_model.required_features
            ]
            udf_kwargs = {
                feat_name: parent_op.values[feat_name]
                for feat_name in required_features
            }
            if isinstance(pruned_udf_wrapper, Callable):
                new_values[name] = pruned_udf_wrapper(**udf_kwargs)
        else:
            new_values[name] = value

    if not pruned_udf_wrapper:
        return _

    new_project = ops.Project(parent_op.parent, new_values)

    subs = {
        ops.Field(parent_op, k): ops.Field(new_project, k) for k in parent_op.values
    }
    new_predicates = tuple(p.replace(subs, filter=ops.Value) for p in _.predicates)
    return ops.Filter(parent=new_project, predicates=new_predicates)


def rewrite_quickgrove_expr(expr: ir.Table) -> ir.Table:
    """Rewrite an Ibis expression by pruning quickgrove models based on filter conditions."""

    op = expr.op()
    new_op = op.replace(prune_quickgrove_model)
    return new_op.to_expr()


__all__ = [
    "make_quickgrove_udf",
    "rewrite_quickgrove_expr",
]



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Iterable, Mapping

import pyarrow as pa
import toolz

from xorq.common.utils.rbr_utils import (
    make_filtered_reader,
)
from xorq.expr.relations import FlightUDXF
from xorq.flight.action import (
    DropTableAction,
    DropViewAction,
    GetExchangeAction,
    GetSchemaQueryAction,
    ListExchangesAction,
    ListTablesAction,
    ReadParquetAction,
    TableInfoAction,
    VersionAction,
)
from xorq.flight.client import FlightClient
from xorq.flight.exchanger import UnboundExprExchanger, make_udxf
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.expr import schema as sch
from xorq.vendor.ibis.expr import types as ir
from xorq.vendor.ibis.util import gen_name


if TYPE_CHECKING:
    import pandas as pd


class Backend(SQLBackend):
    @property
    def name(self):
        return "xorq_flight"

    @property
    def version(self) -> str:
        return self.con.do_action(VersionAction.name, options=self.con._options)[0]

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        import pandas as pd

        if isinstance(obj, pd.DataFrame):
            obj = pa.Table.from_pandas(obj)
        if isinstance(obj, pa.Table):
            self.con.upload_table(name, obj)
            return self.table(name)
        if isinstance(obj, pa.RecordBatchReader):
            self.con.upload_batches(name, obj)
            return self.table(name)
        raise TypeError(f"Unsupported type for create_table: {type(obj)}")

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        return self.con.do_action_one(
            GetSchemaQueryAction.name, action_body=query, options=self.con._options
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.con = None

    def do_connect(
        self,
        host="localhost",
        port=8815,
        username=None,
        password=None,
        cert_chain=None,
        private_key=None,
        tls_root_certs=None,
    ) -> None:
        self.con = FlightClient(
            host=host,
            port=port,
            username=username,
            password=password,
            cert_chain=cert_chain,
            private_key=private_key,
            tls_root_certs=tls_root_certs,
        )

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        args = {
            "table_name": table_name,
            "catalog": catalog,
            "database": database,
        }
        return self.con.do_action_one(
            TableInfoAction.name, action_body=args, options=self.con._options
        )

    def read_in_memory(
        self,
        source: pd.DataFrame | pa.Table | pa.RecordBatchReader,
        table_name: str | None = None,
    ) -> ir.Table:
        import pandas as pd

        table_name = table_name or util.gen_name("read_in_memory")

        if isinstance(source, pa.Table):
            self.con.upload_table(table_name, source)
        elif isinstance(source, pd.DataFrame):
            self.con.upload_batches(
                table_name, pa.RecordBatchReader.from_stream(source)
            )
        elif isinstance(source, pa.RecordBatchReader):
            self.con.upload_batches(table_name, source)
        return self.table(table_name)

    def read_parquet(
        self,
        source_list: str | Iterable[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        args = {
            "source_list": source_list,
            "table_name": table_name,
        }
        table_name = self.con.do_action_one(
            ReadParquetAction.name, action_body=args, options=self.con._options
        )
        return self.table(table_name)

    def register(
        self,
        source: str | Path | Any,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        import pandas as pd

        if isinstance(source, pd.DataFrame):
            source = pa.Table.from_pandas(source)
            source = pa.RecordBatchReader.from_batches(
                source.schema, source.to_batches()
            )

        return self.read_record_batches(source, table_name=table_name)

    def read_record_batches(self, source, table_name=None):
        table_name = table_name or gen_name("read_record_batches")
        self.con.upload_batches(table_name, source)
        return self.table(table_name)

    def list_tables(
        self, like: str | None = None, database: tuple[str, str] | str | None = None
    ) -> list[str]:
        kwargs = {
            "like": like,
            "database": database,
        }

        return self.con.do_action_one(
            ListTablesAction.name, action_body=kwargs, options=self.con._options
        )

    def drop_table(
        self,
        name: str,
        database: tuple[str, str] | str | None = None,
        force: bool = False,
    ) -> None:
        kwargs = {"name": name, "database": database, "force": force}
        self.con.do_action(
            DropTableAction.name, action_body=kwargs, options=self.con._options
        )

    def drop_view(
        self,
        name: str,
        *,
        database: str | None = None,
        schema: str | None = None,
        force: bool = False,
    ) -> None:
        kwargs = {"name": name, "database": database, "schema": schema, "force": force}
        self.con.do_action(
            DropViewAction.name, action_body=kwargs, options=self.con._options
        )

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 10_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        batches = self.con.execute_batches(
            expr, params=params, limit=limit, chunk_size=chunk_size
        )
        batches = make_filtered_reader(batches)
        return batches

    def get_exchange(self, command):
        exchanger = self.con.do_action_one(GetExchangeAction.name, action_body=command)

        if isinstance(exchanger, UnboundExprExchanger):
            exchanger = make_udxf(
                exchanger.exchange_f,  # process_df no wraps
                exchanger.schema_in_required,  # maybe_schema_in
                exchanger.calc_schema_out,  # maybe_schema_out
                name=exchanger.command,  # name
                description=exchanger.description,  # description
                command=exchanger.command,  # command
                do_wraps=False,  #
            )

        @toolz.curry
        def flight_udxf(
            expr,
            con=None,
            name=None,
            inner_name=None,
            **kwargs,
        ):
            @contextmanager
            def make_server():
                try:
                    yield SimpleNamespace(client=self.con)
                finally:
                    pass

            return (
                FlightUDXF.from_expr(
                    input_expr=expr,
                    udxf=exchanger,
                    make_server=make_server,
                    **kwargs,
                )
                .to_expr()
                .into_backend(con=con or expr._find_backend(), name=name)
            )

        return flight_udxf

    def list_exchanges(self):
        return self.con.do_action_one(ListExchangesAction.name)



from __future__ import annotations

from warnings import catch_warnings

import numpy as np
import pytest
from pytest import param

from xorq.backends.pandas import Backend
from xorq.backends.pandas.kernels import sql_like_to_regex
from xorq.tests.util import assert_series_equal
from xorq.vendor import ibis


@pytest.mark.parametrize(
    ("case_func", "expected_func"),
    [
        param(
            lambda s: s.length(),
            lambda s: s.str.len().astype("int32"),
            id="length",
        ),
        param(lambda s: s.substr(1, 2), lambda s: s.str[1:3], id="substr"),
        param(lambda s: s[1:3], lambda s: s.str[1:3], id="slice"),
        param(
            lambda s: s[s.length() - 1 :],
            lambda s: s.str[-1:],
            id="expr_slice_begin",
        ),
        param(lambda s: s[: s.length()], lambda s: s, id="expr_slice_end"),
        param(
            lambda s: s[s.length() - 2 : s.length() - 1],
            lambda s: s.str[-2:-1],
            id="expr_slice_begin_end",
        ),
        param(lambda s: s.strip(), lambda s: s.str.strip(), id="strip"),
        param(lambda s: s.lstrip(), lambda s: s.str.lstrip(), id="lstrip"),
        param(lambda s: s.rstrip(), lambda s: s.str.rstrip(), id="rstrip"),
        param(
            lambda s: s.lpad(3, "a"),
            lambda s: s.str.pad(3, side="left", fillchar="a"),
            id="lpad",
        ),
        param(
            lambda s: s.rpad(3, "b"),
            lambda s: s.str.pad(3, side="right", fillchar="b"),
            id="rpad",
        ),
        param(lambda s: s.reverse(), lambda s: s.str[::-1], id="reverse"),
        param(lambda s: s.lower(), lambda s: s.str.lower(), id="lower"),
        param(lambda s: s.upper(), lambda s: s.str.upper(), id="upper"),
        param(lambda s: s.repeat(2), lambda s: s * 2, id="repeat"),
        param(
            lambda s: s.contains("a"),
            lambda s: s.str.contains("a", regex=False),
            id="contains",
        ),
        param(
            lambda s: ~(s.contains("a")),
            lambda s: ~s.str.contains("a", regex=False),
            id="not_contains",
        ),
        param(
            lambda s: s.like("a"),
            lambda s: s.str.contains("^a$", regex=True),
            id="like",
        ),
        param(
            lambda s: s.re_search("(ab)+"),
            lambda s: s.str.contains("(?:ab)+", regex=True),
            id="re_search",
        ),
        param(
            lambda s: s.re_search("(ab)+") | s.re_search("d{1,2}ee"),
            lambda s: (
                s.str.contains("(?:ab)+", regex=True) | s.str.contains("d{1,2}ee")
            ),
            id="re_search_or",
        ),
        param(
            lambda s: s + s.rpad(3, "a"),
            lambda s: s + s.str.pad(3, side="right", fillchar="a"),
            id="rpad2",
        ),
        param(
            lambda s: s.split(" "),
            lambda s: s.apply(lambda x: np.array(x.split(" "))),
            id="split_spaces",
        ),
    ],
)
def test_string_ops(t, df, case_func, expected_func):
    # ignore matching UserWarnings
    with catch_warnings(record=True):
        expr = case_func(t.strings_with_space)
        result = expr.execute()
        series = expected_func(df.strings_with_space)
        assert_series_equal(result, series, check_names=False)


@pytest.mark.parametrize(
    ("pattern", "expected"),
    [
        ("%abc", ".*abc"),
        ("abc%", "abc.*"),
        ("6%", "6.*"),
        ("%6%", ".*6.*"),
        ("^%6", "%6"),
        ("6^%", "6%"),
        ("6^%%", "6%.*"),
        ("^%%6", "%.*6"),
        ("^%^%6", "%%6"),
        ("6^%^%", "6%%"),
        ("6_", "6."),
        ("_6_", ".6."),
        ("^_6", "_6"),
        ("6^_", "6_"),
        ("6^__", "6_."),
        ("^__6", "_.6"),
        ("^_^_6", "__6"),
        ("6^_^_", "6__"),
        ("6%_^%_", "6.*.%."),
        ("6_^%%_", "6.%.*."),
        ("_^%%_%_^%_%_^%^__^%%^_^%%6%_", ".%.*..*.%..*.%_.%.*_%.*6.*."),
    ],
)
def test_sql_like_to_regex(pattern, expected):
    result = sql_like_to_regex(pattern, escape="^")
    assert result == f"^{expected}$"


@pytest.mark.parametrize(
    ("from_func", "to_func", "from_str", "to_str"),
    [
        param(
            lambda s: s.translate_from_strings,
            lambda s: s.translate_to_strings,
            "rmzabcghj",
            "lnsovkjfr",
            id="from_series_to_series",
        ),
        param(
            lambda s: "abc",
            lambda s: s.translate_to_strings,
            "abc",
            "ovk",
            id="from_string_to_series",
        ),
        param(
            lambda s: s.translate_from_strings,
            lambda s: "ovk",
            "abcg",
            "ovko",
            id="from_series_to_string",
        ),
    ],
)
def test_translate(
    t, df, from_func: callable, to_func: callable, from_str: str, to_str: str
):
    result = t.strings_with_space.translate(from_func(t), to_func(t)).execute()
    table = str.maketrans(from_str, to_str)
    series = df.strings_with_space.str.translate(table)
    assert_series_equal(result, series, check_names=False)


def test_string_repeat(t):
    int_col = t.plain_int64
    int_lit = ibis.literal(3)
    string_col = t.strings_with_space
    string_lit = ibis.literal("abc")

    expr1 = string_col.repeat(int_col)
    expr2 = string_col.repeat(int_lit)
    expr3 = string_lit.repeat(int_col)
    expr4 = string_lit.repeat(int_lit)

    con = Backend()
    con.execute(expr1)
    con.execute(expr2)
    con.execute(expr3)
    con.execute(expr4)

    # TODO(kszucs): add assertions or rather parametrize the tests above



from typing import Any, Dict

import toolz

import xorq.expr.datatypes as dt
import xorq.expr.udf as udf
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.rules as rlz
from xorq.expr.relations import FlightExpr, FlightUDXF
from xorq.ibis_yaml.common import (
    RefEnum,
    RegistryEnum,
    TranslationContext,
    deserialize_callable,
    register_from_yaml_handler,
    serialize_callable,
    translate_from_yaml,
    translate_to_yaml,
)
from xorq.ibis_yaml.utils import freeze
from xorq.vendor import ibis
from xorq.vendor.ibis.common.annotations import Argument


@translate_to_yaml.register(ops.ScalarUDF)
def _scalar_udf_to_yaml(op: ops.ScalarUDF, compiler: Any) -> dict:
    input_type = getattr(op.__class__, "__input_type__", None)

    if input_type not in [ops.udf.InputType.BUILTIN, ops.udf.InputType.PYARROW]:
        raise NotImplementedError(
            f"Translation of UDFs with input type {getattr(op.__class__, '__input_type__', None)} is not supported"
        )
    arg_names = [
        name
        for name in dir(op)
        if not name.startswith("__") and name not in op.__class__.__slots__
    ]

    return freeze(
        {
            "op": "ScalarUDF",
            "func_name": op.__func_name__,
            "input_type": str(input_type),
            "args": [compiler.translate_to_yaml(arg) for arg in op.args],
            "type": translate_to_yaml(op.dtype, None),
            "pickle": serialize_callable(op.__func__),
            "module": op.__module__,
            "class_name": op.__class__.__name__,
            "arg_names": arg_names,
        }
    )


@register_from_yaml_handler("ScalarUDF")
def _scalar_udf_from_yaml(yaml_dict: dict, compiler: any) -> any:
    encoded_fn = yaml_dict.get("pickle")
    if not encoded_fn:
        raise ValueError("Missing pickle data for ScalarUDF")
    fn = deserialize_callable(encoded_fn)

    args_yaml = yaml_dict.get("args", [])
    args = [translate_from_yaml(arg, compiler) for arg in args_yaml]
    if not args:
        raise ValueError("ScalarUDF requires at least one argument")

    input_type_str = yaml_dict.get("input_type")

    if input_type_str == "InputType.BUILTIN":
        input_type = ops.udf.InputType.BUILTIN
    elif input_type_str == "InputType.PYARROW":
        input_type = ops.udf.InputType.PYARROW
    else:
        raise ValueError(f"Unsupported input type: {input_type_str}")

    dtype = dt.dtype(yaml_dict["type"]["type"])
    class_name = yaml_dict.get("class_name", yaml_dict["func_name"])

    schema = {}
    for i, arg_yaml in enumerate(args_yaml):
        arg_name = f"arg{i}"

        if RefEnum.node_ref in arg_yaml and (
            node := compiler.definitions[RegistryEnum.nodes].get(
                arg_yaml[RefEnum.node_ref]
            )
        ):
            if node.get("op") == "Field" and "name" in node:
                arg_name = node["name"]

        schema[arg_name] = args[i].type()

    fields = {}
    for name, typ in schema.items():
        fields[name] = Argument(pattern=rlz.ValueOf(typ), typehint=typ)

    meta = {
        "dtype": dtype,
        "__input_type__": input_type,
        "__func__": udf.property_wrap_fn(fn),
        "__config__": {"volatility": "immutable"},
        "__udf_namespace__": None,
        "__module__": yaml_dict.get("module", "__main__"),
        "__func_name__": yaml_dict["func_name"],
    }

    kwds = {**fields, **meta}
    node = type(
        class_name,
        (ops.ScalarUDF,),
        kwds,
    )

    return node(*args).to_expr()


@translate_to_yaml.register(dict)
def _dict_to_yaml(dct: dict, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "dict",
        }
        | {key: context.translate_to_yaml(value) for key, value in dct.items()}
    )


@register_from_yaml_handler("dict")
def _dict_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    dct = {
        key: context.translate_from_yaml(value)
        for key, value in toolz.dissoc(yaml_dict, "op").items()
    }
    return dct


@translate_to_yaml.register(ops.Namespace)
def _namespace_to_yaml(ns: ops.Namespace, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "Namespace",
            "dict": dict(zip(ns.argnames, ns.args)),
        }
    )


@register_from_yaml_handler("Namespace")
def _namespace_from_yaml(yaml_dict: dict, compiler: any) -> any:
    return ops.Namespace(**yaml_dict["dict"])


@translate_to_yaml.register(ops.udf.InputType)
def _inputtype_to_yaml(it: ops.udf.InputType, context: TranslationContext) -> dict:
    return freeze(
        {
            "op": "InputType",
            "name": it.name,
        }
    )


@register_from_yaml_handler("InputType")
def _inputtype_from_yaml(yaml_dict: dict, context: TranslationContext) -> any:
    return getattr(ops.udf.InputType, yaml_dict["name"])


def require_input_types(input_types, op):
    if (input_type := getattr(op.__class__, "__input_type__", None)) not in input_types:
        raise NotImplementedError(
            f"Translation of UDFs with input type {input_type} is not supported"
            f"\n\tsupported input types: {input_types}"
        )


def make_op_kwargs(op):
    argnames = op.argnames
    if argnames and argnames[-1] == "where":
        (*argnames, _) = argnames
    kwargs = dict(zip(argnames, op.args))
    return kwargs


def kwargs_to_schema(kwargs):
    schema = ibis.schema({argname: arg.type() for argname, arg in kwargs.items()})
    return schema


@translate_to_yaml.register(ops.udf.AggUDF)
def _aggudf_to_yaml(op: ops.udf.AggUDF, compiler: Any) -> dict:
    require_input_types((ops.udf.InputType.PYARROW,), op)
    kwargs = make_op_kwargs(op)
    meta = {
        name: getattr(op, name)
        for name in (
            "dtype",
            "__config__",
            "__input_type__",
            "__udf_namespace__",
            "__module__",
            "__func_name__",
        )
    }
    return freeze(
        {
            "op": "AggUDF",
            "class_name": op.__class__.__name__,
            "kwargs": compiler.translate_to_yaml(freeze(kwargs)),
            "meta": compiler.translate_to_yaml(freeze(meta)),
        }
    )


@register_from_yaml_handler("AggUDF")
def _aggudf_from_yaml(yaml_dict: dict, compiler: any) -> any:
    (kwargs, meta) = (
        translate_from_yaml(yaml_dict[name], compiler) for name in ("kwargs", "meta")
    )
    fields = {
        argname: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (argname, typ) in ((argname, arg.type()) for argname, arg in kwargs.items())
    }
    #
    class_name = yaml_dict["class_name"]
    bases = (ops.udf.AggUDF,)
    kwds = fields | meta

    config = meta["__config__"]
    if "fn" in config:
        kwds["__func__"] = udf.make_dunder_func(config["fn"], kwargs)
    elif {
        "evaluate",
        "evaluate_all",
        "evaluate_all_with_rank",
    } & config.keys():
        kwds["__func__"] = property(fget=toolz.functoolz.return_none)

    node = type(
        class_name,
        bases,
        kwds,
    )
    return node(**kwargs).to_expr()


@translate_to_yaml.register(udf.ExprScalarUDF)
def _exprscalarudf_to_yaml(op: udf.ExprScalarUDF, compiler: Any) -> dict:
    require_input_types((ops.udf.InputType.PYARROW,), op)
    kwargs = make_op_kwargs(op)
    meta = {
        name: getattr(op, name)
        for name in (
            "dtype",
            "__input_type__",
            "__config__",
            "__udf_namespace__",
            "__module__",
            "__func_name__",
        )
    }
    return freeze(
        {
            "op": udf.ExprScalarUDF.__name__,
            "class_name": op.__class__.__name__,
            "kwargs": compiler.translate_to_yaml(freeze(kwargs)),
            "meta": compiler.translate_to_yaml(freeze(meta)),
        }
    )


@register_from_yaml_handler(udf.ExprScalarUDF.__name__)
def _aggudf_from_yaml(yaml_dict: dict, compiler: any) -> any:
    (kwargs, meta) = (
        translate_from_yaml(yaml_dict[name], compiler) for name in ("kwargs", "meta")
    )
    fields = {
        argname: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
        for (argname, typ) in ((argname, arg.type()) for argname, arg in kwargs.items())
    }
    #
    class_name = yaml_dict["class_name"]
    bases = (udf.ExprScalarUDF,)
    kwds = (
        fields
        | meta
        | {
            "__func__": udf.make_expr_scalar_udf_dunder_func(
                meta["__config__"]["fn"], kwargs_to_schema(kwargs), meta["dtype"]
            )
        }
    )
    node = type(
        class_name,
        bases,
        kwds,
    )
    return node(**kwargs).to_expr()


@translate_to_yaml.register(FlightExpr)
def flight_expr_to_yaml(op: FlightExpr, context: any) -> dict:
    input_expr_yaml = context.translate_to_yaml(op.input_expr)
    unbound_expr_yaml = context.translate_to_yaml(op.unbound_expr)
    make_server_pickle = serialize_callable(op.make_server)
    make_connection_pickle = serialize_callable(op.make_connection)
    return freeze(
        {
            "op": "FlightExpr",
            "name": op.name,
            "input_expr": input_expr_yaml,
            "unbound_expr": unbound_expr_yaml,
            "make_server": make_server_pickle,
            "make_connection": make_connection_pickle,
            "do_instrument_reader": op.do_instrument_reader,
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("FlightExpr")
def flight_expr_from_yaml(yaml_dict: Dict, context: Any) -> Any:
    name = yaml_dict.get("name")
    input_expr_yaml = yaml_dict.get("input_expr")
    unbound_expr_yaml = yaml_dict.get("unbound_expr")
    make_server_pickle = yaml_dict.get("make_server")
    make_connection_pickle = yaml_dict.get("make_connection")
    do_instrument_reader = yaml_dict.get("do_instrument_reader", False)

    input_expr = context.translate_from_yaml(input_expr_yaml)
    unbound_expr = context.translate_from_yaml(unbound_expr_yaml)

    make_server = (
        deserialize_callable(make_server_pickle) if make_server_pickle else None
    )
    make_connection = (
        deserialize_callable(make_connection_pickle) if make_connection_pickle else None
    )

    return FlightExpr.from_exprs(
        input_expr=input_expr,
        unbound_expr=unbound_expr,
        make_server=make_server,
        make_connection=make_connection,
        name=name,
        do_instrument_reader=do_instrument_reader,
    ).to_expr()


@translate_to_yaml.register(FlightUDXF)
def flight_udxf_to_yaml(op: FlightUDXF, context: any) -> dict:
    input_expr_yaml = context.translate_to_yaml(op.input_expr)
    udxf_pickle = serialize_callable(op.udxf)
    make_server_pickle = serialize_callable(op.make_server)
    make_connection_pickle = serialize_callable(op.make_connection)

    return freeze(
        {
            "op": "FlightUDXF",
            "name": op.name,
            "input_expr": input_expr_yaml,
            "udxf": udxf_pickle,
            "make_server": make_server_pickle,
            "make_connection": make_connection_pickle,
            "do_instrument_reader": op.do_instrument_reader,
        }
        | context.registry.register_schema(op.schema)
    )


@register_from_yaml_handler("FlightUDXF")
def flight_udxf_from_yaml(yaml_dict: Dict, context: Any) -> Any:
    name = yaml_dict.get("name")
    input_expr_yaml = yaml_dict.get("input_expr")
    udxf_pickle = yaml_dict.get("udxf")
    make_server_pickle = yaml_dict.get("make_server")
    make_connection_pickle = yaml_dict.get("make_connection")
    do_instrument_reader = yaml_dict.get("do_instrument_reader", False)

    input_expr = context.translate_from_yaml(input_expr_yaml)
    udxf = deserialize_callable(udxf_pickle)
    make_server = (
        deserialize_callable(make_server_pickle) if make_server_pickle else None
    )
    make_connection = (
        deserialize_callable(make_connection_pickle) if make_connection_pickle else None
    )

    return FlightUDXF.from_expr(
        input_expr=input_expr,
        udxf=udxf,
        make_server=make_server,
        make_connection=make_connection,
        name=name,
        do_instrument_reader=do_instrument_reader,
    ).to_expr()



import pickle

import pytest

import xorq.api as xo


@pytest.mark.parametrize(
    "make_connection",
    (
        lambda: xo.postgres.connect_env(),
        lambda: xo.postgres.connect_examples(),
    ),
)
@pytest.mark.slow(level=1)
def test_remote_reconnect(make_connection):
    con = make_connection()
    expected = con.list_tables()
    actual = pickle.loads(pickle.dumps(con)).list_tables()
    assert expected  # smoke test that we have something meaningful to test
    assert actual == expected



"""Subquery operations."""

from __future__ import annotations

from public import public

import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.common.exceptions import IntegrityError
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.expr.operations.core import Value
from xorq.vendor.ibis.expr.operations.relations import Relation  # noqa: TCH001


@public
class Subquery(Value):
    """Base class for subquery operations."""

    rel: Relation

    @attribute
    def relations(self):
        return frozenset()


@public
class ExistsSubquery(Subquery):
    """Check if a subquery returns any rows."""

    dtype = dt.boolean
    shape = ds.columnar


@public
class ScalarSubquery(Subquery):
    """A subquery that returns a single scalar value."""

    shape = ds.scalar

    def __init__(self, rel):
        if len(rel.schema) != 1:
            raise IntegrityError(
                "Relation passed to ScalarSubquery() must have exactly one "
                f"column, got {len(rel.schema)}"
            )
        super().__init__(rel=rel)

    @attribute
    def value(self):
        (value,) = self.rel.values.values()
        return value

    @attribute
    def dtype(self):
        return self.value.dtype


@public
class InSubquery(Subquery):
    """Check if a value is in the result of a subquery."""

    needle: Value

    dtype = dt.boolean
    shape = rlz.shape_like("needle")

    def __init__(self, rel, needle):
        if len(rel.schema) != 1:
            raise IntegrityError(
                "Relation passed to InSubquery() must have exactly one "
                f"column, got {len(rel.schema)}"
            )
        (value,) = rel.values.values()
        if not rlz.comparable(value, needle):
            raise IntegrityError(f"{needle!r} is not comparable to {value!r}")
        super().__init__(rel=rel, needle=needle)

    @attribute
    def value(self):
        (value,) = self.rel.values.values()
        return value

    @attribute
    def relations(self):
        return self.needle.relations



from __future__ import annotations

from functools import lru_cache, singledispatch
from itertools import count
from typing import Any, Callable, Tuple

import dask.base
from attrs import evolve, field, frozen
from attrs.validators import instance_of

import xorq.expr.relations as rel
import xorq.expr.udf as udf
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.graph_utils import (
    bfs,
    gen_children_of,
    to_node,
)
from xorq.vendor.ibis.expr.operations.core import Node


__all__ = [
    "build_column_trees",
    "build_tree",
]


@frozen
class TextTree:
    """Plain-text tree for displaying lineage."""

    label: str = field(validator=instance_of(str))
    children: Tuple["TextTree", ...] = field(
        factory=tuple, validator=instance_of(tuple)
    )

    def _lines(
        self, prefix: str = "", is_last: bool = True, is_root: bool = True
    ) -> tuple[str, ...]:
        if is_root:
            line = self.label
            child_prefix = ""
        else:
            connector = "└── " if is_last else "├── "
            line = prefix + connector + self.label
            child_prefix = prefix + ("    " if is_last else "│   ")
        return (line,) + tuple(
            grandchild_line
            for i, child in enumerate(self.children)
            for grandchild_line in child._lines(
                child_prefix, i == len(self.children) - 1, False
            )
        )

    def __str__(self) -> str:
        return "\n".join(self._lines())


@frozen
class GenericNode:
    op: Node = field(validator=instance_of(Node))
    children: Tuple["GenericNode", ...] = field(
        factory=tuple, validator=instance_of(tuple)
    )

    def map_children(
        self, fn: Callable[["GenericNode"], "GenericNode"]
    ) -> "GenericNode":
        return evolve(self, children=tuple(fn(c) for c in self.children))

    def clone(self, **changes: Any) -> "GenericNode":
        return evolve(self, **changes)


def _build_column_tree(node: Node) -> GenericNode:
    graph, _ = bfs(node).toposort()
    results: dict[Node, GenericNode] = {}

    for n in graph:
        match n:
            case ops.Field(rel=ops.Project(values=values)) as field_node:
                # include the field and follow it into its mapped expression
                child = results[to_node(values[field_node.name])]
                results[n] = GenericNode(op=field_node, children=(child,))

            case ops.Field() as field_node:
                children = tuple(results[c] for c in gen_children_of(field_node))
                results[n] = GenericNode(op=field_node, children=children)

            case ops.Project() as proj:
                # Project is transparent: resolve to its parent's GenericNode
                results[n] = results[to_node(proj.parent)]

            case _:
                children = tuple(results[c] for c in gen_children_of(n))
                results[n] = GenericNode(op=n, children=children)

    return results[node]


def build_column_trees(expr: Any) -> dict[str, GenericNode]:
    """Builds a lineage tree for each column in the expression."""
    op = to_node(expr)
    cols = getattr(op, "values", None) or getattr(op, "fields", {})
    return {k: _build_column_tree(to_node(v)) for k, v in cols.items()}


@singledispatch
def format_node(node: Node) -> str:
    return node.__class__.__name__


@format_node.register
def _(node: ops.Field) -> str:
    return f"Field:{node.name}"


@format_node.register
def _(node: rel.RemoteTable) -> str:
    return f"RemoteTable:{node.name}"


@format_node.register
def _(node: rel.CachedNode) -> str:
    store = getattr(node.cache, "kind", "cache")
    return f"Cache[{store}] {getattr(node, 'name', '')}"


@format_node.register
def _(node: rel.FlightExpr) -> str:
    return f"FlightExpr ({node.input_expr})"


@format_node.register
def _(node: udf.ExprScalarUDF) -> str:
    return "ExprScalarUDF"


@format_node.register
def _(node: ops.WindowFunction) -> str:
    parts = []
    if node.order_by:
        parts.append(f"order_by: {node.order_by}")
    if node.group_by:
        parts.append(f"group_by: {node.group_by}")
    if node.start:
        parts.append(f"start: {node.start}")
    if node.end:
        parts.append(f"end: {node.end}")

    if parts:
        details = "\n ".join(parts)
        return f"WindowFunction:\n {details}"
    return "WindowFunction"


@format_node.register
def _(node: ops.Literal) -> str:
    return f"Literal: {node.value}"


@lru_cache
def _token_node(g: GenericNode) -> str:
    """unique token for the node based on its operation and children (useful in
    deduplication)"""
    op = g.op
    return dask.base.tokenize(
        (
            getattr(op, "name", None),  # is name always set?
            getattr(op, "schema", None),
            tuple(_token_node(c) for c in g.children),
        )
    )


def build_tree(
    node: GenericNode,
    *,
    dedup: bool = True,
    max_depth: int | None = None,
) -> TextTree:
    seen: dict[str, int] = {}
    seq = count(1)

    def _to_tree(g: GenericNode, depth: int) -> TextTree:
        if max_depth is not None and depth > max_depth:
            return TextTree("…")

        digest = _token_node(g) if dedup else None
        if digest is not None and digest in seen:
            ref = seen[digest]
            return TextTree(f"↻ see #{ref}")

        ref = next(seq)
        if digest is not None:
            seen[digest] = ref

        label = format_node(g.op)
        if dedup:
            label += f" #{ref}"
        children = tuple(_to_tree(child, depth + 1) for child in g.children)
        return TextTree(label, children=children)

    return _to_tree(node, 0)



from __future__ import annotations

import datetime
import functools
from abc import (
    abstractmethod,
)
from pathlib import (
    Path,
)

from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
    optional,
)

from xorq.common.utils.caching_utils import (
    get_xorq_cache_dir,
)
from xorq.common.utils.defer_utils import (
    deferred_read_parquet,
)
from xorq.common.utils.func_utils import (
    if_not_none,
)
from xorq.config import _backend_init, options
from xorq.vendor import ibis


@frozen
class CacheStorage:
    @abstractmethod
    def exists(self, key):
        pass

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def put(self, key, value):
        pass

    @abstractmethod
    def drop(self, key):
        pass


@frozen
class ParquetStorage(CacheStorage):
    source = field(
        validator=instance_of(ibis.backends.BaseBackend),
        factory=_backend_init,
    )
    relative_path = field(
        validator=instance_of(Path),
        factory=functools.partial(options.get, "cache.default_relative_path"),
        converter=Path,
    )
    base_path = field(
        validator=optional(instance_of(Path)),
        default=None,
        converter=if_not_none(Path),
    )

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(
            self.source,
            self.relative_path,
            self.base_path,
            caller="normalize_parquet_storage",
        )

    def __attrs_post_init__(self):
        self.path.mkdir(exist_ok=True, parents=True)

    @property
    def path(self):
        return (self.base_path or get_xorq_cache_dir()).joinpath(self.relative_path)

    def get_path(self, key):
        return self.path.joinpath(key + ".parquet")

    def exists(self, key):
        return self.get_path(key).exists()

    def get(self, key):
        op = deferred_read_parquet(
            path=self.get_path(key),
            con=self.source,
            table_name=key,
        ).op()
        return op

    def put(self, key, value):
        path = self.get_path(key)
        # move from temp location upon success to prevent empty files on failure
        tmp_path = path.with_name(path.name + ".tmp")
        value.to_expr().to_parquet(tmp_path)
        tmp_path.rename(path)
        return self.get(key)

    def drop(self, key):
        path = self.get_path(key)
        path.unlink()


@frozen
class ParquetTTLStorage(ParquetStorage):
    ttl = field(
        validator=instance_of(datetime.timedelta), default=datetime.timedelta(days=1)
    )

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(
            self.source,
            self.relative_path,
            self.base_path,
            self.ttl,
            caller="normalize_parquet_ttl_storage",
        )

    def exists(self, key):
        path = self.get_path(key)
        return path.exists() and self.satisfies_ttl(path)

    def satisfies_ttl(self, path):
        delta = datetime.datetime.now() - datetime.datetime.fromtimestamp(
            path.stat().st_mtime
        )
        return delta < self.ttl


@frozen
class SourceStorage(CacheStorage):
    source = field(
        validator=instance_of(ibis.backends.BaseBackend),
        factory=_backend_init,
    )

    def __dask_tokenize__(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import (
            normalize_seq_with_caller,
        )

        return normalize_seq_with_caller(self.source, caller="normalize_source_storage")

    def exists(self, key):
        return key in self.source.tables

    def get(self, key):
        return self.source.table(key).op()

    def put(self, key, value):
        def is_remote(value):
            name = value.to_expr()._find_backend().name
            # FIXME: add pyiceberg, trino
            return name in ("postgres", "snowflake")

        def is_single_backend(storage, value):
            from xorq.common.utils.graph_utils import find_all_sources

            return (storage.source,) == find_all_sources(value.to_expr())

        if is_remote(value):
            if is_single_backend(self, value):
                from xorq.expr.api import _transform_expr

                # must transform for Read ops: create_table expects a vanilla ibis expr
                (transformed, _) = _transform_expr(value.to_expr())
                self.source.create_table(key, transformed)
            else:
                assert hasattr(self.source, "read_record_batches")
                # read_record_batches will create durable table in out-of-core fashion
                # works for snowflake and postgres
                self.source.read_record_batches(
                    value.to_expr().to_pyarrow_batches(),
                    key,
                )
        else:
            self.source.create_table(key, value.to_expr().to_pyarrow())
        return self.get(key)

    def drop(self, key):
        self.source.drop_table(key)



from __future__ import annotations

from typing import Any

from streamlit.connections import BaseConnection
from streamlit.runtime.caching import cache_data

from xorq.vendor import ibis
from xorq.vendor.ibis.backends import BaseBackend


__all__ = ["IbisConnection"]


class IbisConnection(BaseConnection[BaseBackend]):
    def _connect(self, **kwargs: Any) -> BaseBackend:
        """Connect to the backend and return a client object.

        This method is invoked when `st.connection` is called and pulls
        information from streamlit secrets. `_connect` is part of the streamlit
        connection API to be implemented by developers of specific connection
        types.

        Here's an example not-so-secret configuration:

        ```toml
        [connections.ch]
        url = "clickhouse://play:clickhouse@play.clickhouse.com:9440/?secure=1"
        ```

        Alternatively, you can specify individual arguments under a connection
        whose name matches the backend type. For example:

        ```toml
        [connections.clickhouse]
        user = "play"
        password = "clickhouse"
        host = "play.clickhouse.com"
        port = 9440
        secure = 1
        ```

        This file can be placed at `~/.streamlit/secrets.toml`.

        You can then connect to the backend using:

        ```python
        import streamlit as st

        from xorq.vendor.ibis.streamlit import IbisConnection

        con = st.connection("ch", type=IbisConnection)

        # Now you can use `con` as if it were an ibis backend
        con.list_tables()
        ```
        """
        # add secrets config to kwargs
        kwargs.update(self._secrets.to_dict())

        # pop out the connection string
        url = kwargs.pop("url", None)

        # if there's no url parameter, use the connection name to get connection arguments
        if url is None:
            return getattr(ibis, self._connection_name).connect(**kwargs)

        return ibis.connect(url, **kwargs)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._instance, name)

    @cache_data
    def list_tables(_self, *args, **kwargs) -> list[str]:
        return _self._instance.list_tables(*args, **kwargs)



from xorq.backends.pyiceberg import (
    Backend as PyIcebergBackend,
)
from xorq.common.utils.env_utils import EnvConfigable, env_templates_dir


PyIcebergConfig = EnvConfigable.subclass_from_env_file(
    env_templates_dir.joinpath(".env.pyiceberg.template")
)
pyiceberg_config = PyIcebergConfig.from_env()


def make_connection_defaults():
    return {
        "uri": pyiceberg_config["ICEBERG_URI"],
        "warehouse_path": pyiceberg_config["ICEBERG_WAREHOUSE_PATH"],
        "namespace": pyiceberg_config["ICEBERG_NAMESPACE"],
        "catalog_name": pyiceberg_config["ICEBERG_CATALOG_NAME"],
        "catalog_type": pyiceberg_config["ICEBERG_CATALOG_TYPE"],
    }


def make_connection(**kwargs):
    con = PyIcebergBackend()
    con = con.connect(
        **{
            **make_connection_defaults(),
            **kwargs,
        }
    )
    return con


def get_iceberg_snapshots_ids(dt):
    database = dt.source.namespace
    catalog = dt.source.catalog
    table_name = dt.name

    table = catalog.load_table(f"{database}.{table_name}")

    return tuple(snapshot.snapshot_id for snapshot in table.metadata.snapshots)



from __future__ import annotations

import collections.abc
import functools
import inspect
import operator
from abc import abstractmethod
from collections.abc import Callable
from typing import Any, TypeVar, overload

from xorq.vendor.ibis.common.bases import (
    Final,
    FrozenSlotted,
    Hashable,
    Immutable,
    Slotted,
)
from xorq.vendor.ibis.common.collections import FrozenDict
from xorq.vendor.ibis.common.typing import Coercible, CoercionError
from xorq.vendor.ibis.util import PseudoHashable


class Resolver(Coercible, Hashable):
    """Specification about constructing a value given a context.

    The context is a dictionary that contains all the captured values and
    information relevant for the builder.

    The builder is used in the right hand side of the replace pattern:
    `Replace(pattern, builder)`. Semantically when a match occurs for the
    replace pattern, the builder is called with the context and the result
    of the builder is used as the replacement value.
    """

    @abstractmethod
    def resolve(self, context: dict):
        """Construct a new object from the context.

        Parameters
        ----------
        context
            A dictionary containing all the captured values and information
            relevant for the deferred.

        Returns
        -------
        The constructed object.

        """

    @abstractmethod
    def __eq__(self, other: Resolver) -> bool: ...

    @classmethod
    def __coerce__(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, Deferred):
            return value._resolver
        else:
            raise CoercionError(
                f"Cannot coerce {type(value).__name__!r} to {cls.__name__!r}"
            )


class Deferred(Slotted, Immutable, Final):
    """The user facing wrapper object providing syntactic sugar for deferreds.

    Provides a natural-like syntax for constructing deferred expressions by
    overloading all of the available dunder methods including the equality
    operator.

    Its sole purpose is to provide a nicer syntax for constructing deferred
    expressions, thus it gets unwrapped to the underlying deferred expression
    when used by the rest of the library.

    Parameters
    ----------
    deferred
        The deferred object to provide syntax sugar for.
    repr
        An optional fixed string to use when repr-ing the deferred expression,
        instead of the default. This is useful for complex deferred expressions
        where the arguments don't necessarily make sense to be user facing in
        the repr.

    """

    __slots__ = ("_resolver", "_repr")

    def __init__(self, obj, repr=None):
        super().__init__(_resolver=resolver(obj), _repr=repr)

    # TODO(kszucs): consider to make this method protected
    def resolve(self, _=None, **kwargs):
        context = {"_": _, **kwargs}
        return self._resolver.resolve(context)

    def __repr__(self):
        return repr(self._resolver) if self._repr is None else self._repr

    def __getattr__(self, name):
        return Deferred(Attr(self, name))

    def __iter__(self):
        raise TypeError(f"{self.__class__.__name__!r} object is not iterable")

    def __bool__(self):
        raise TypeError(
            f"The truth value of {self.__class__.__name__} objects is not defined"
        )

    def __getitem__(self, name):
        return Deferred(Item(self, name))

    def __call__(self, *args, **kwargs):
        return Deferred(Call(self, *args, **kwargs))

    def __invert__(self) -> Deferred:
        return Deferred(UnaryOperator(operator.invert, self))

    def __neg__(self) -> Deferred:
        return Deferred(UnaryOperator(operator.neg, self))

    def __add__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.add, self, other))

    def __radd__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.add, other, self))

    def __sub__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.sub, self, other))

    def __rsub__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.sub, other, self))

    def __mul__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.mul, self, other))

    def __rmul__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.mul, other, self))

    def __truediv__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.truediv, self, other))

    def __rtruediv__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.truediv, other, self))

    def __floordiv__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.floordiv, self, other))

    def __rfloordiv__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.floordiv, other, self))

    def __pow__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.pow, self, other))

    def __rpow__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.pow, other, self))

    def __mod__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.mod, self, other))

    def __rmod__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.mod, other, self))

    def __rshift__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.rshift, self, other))

    def __rrshift__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.rshift, other, self))

    def __lshift__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.lshift, self, other))

    def __rlshift__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.lshift, other, self))

    def __eq__(self, other: Any) -> Deferred:  # type: ignore
        return Deferred(BinaryOperator(operator.eq, self, other))

    def __ne__(self, other: Any) -> Deferred:  # type: ignore
        return Deferred(BinaryOperator(operator.ne, self, other))

    def __lt__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.lt, self, other))

    def __le__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.le, self, other))

    def __gt__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.gt, self, other))

    def __ge__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.ge, self, other))

    def __and__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.and_, self, other))

    def __rand__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.and_, other, self))

    def __or__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.or_, self, other))

    def __ror__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.or_, other, self))

    def __xor__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.xor, self, other))

    def __rxor__(self, other: Any) -> Deferred:
        return Deferred(BinaryOperator(operator.xor, other, self))


class Variable(FrozenSlotted, Resolver):
    """Retrieve a value from the context.

    Parameters
    ----------
    name
        The key to retrieve from the state.

    """

    __slots__ = ("name",)
    name: Any

    def __init__(self, name):
        super().__init__(name=name)

    def __repr__(self):
        return str(self.name)

    def resolve(self, context):
        return context[self.name]


class Just(FrozenSlotted, Resolver):
    """Construct exactly the given value.

    Parameters
    ----------
    value
        The value to return when the deferred is called.

    """

    __slots__ = ("value",)
    value: Any

    @classmethod
    def __create__(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, (Deferred, Resolver)):
            raise TypeError(f"{value} cannot be used as a Just value")
        elif isinstance(value, collections.abc.Hashable):
            return super().__create__(value)
        else:
            return JustUnhashable(value)

    def __init__(self, value):
        super().__init__(value=value)

    def __repr__(self):
        obj = self.value
        if hasattr(obj, "__deferred_repr__"):
            return obj.__deferred_repr__()
        elif callable(obj):
            return getattr(obj, "__name__", repr(obj))
        else:
            return repr(obj)

    def resolve(self, context):
        return self.value


class JustUnhashable(FrozenSlotted, Resolver):
    """Construct exactly the given unhashable value.

    Parameters
    ----------
    value
        The value to return when the deferred is called.

    """

    __slots__ = ("value",)

    def __init__(self, value):
        hashable_value = PseudoHashable(value)
        super().__init__(value=hashable_value)

    def __repr__(self):
        obj = self.value.obj
        if hasattr(obj, "__deferred_repr__"):
            return obj.__deferred_repr__()
        elif callable(obj):
            return getattr(obj, "__name__", repr(obj))
        else:
            return repr(obj)

    def resolve(self, context):
        return self.value.obj


class Factory(FrozenSlotted, Resolver):
    """Construct a value by calling a function.

    The function is called with two positional arguments:
    1. the value being matched
    2. the context dictionary

    The function must return the constructed value.

    Parameters
    ----------
    func
        The function to apply.

    """

    __slots__ = ("func",)
    func: Callable

    def __init__(self, func):
        assert callable(func)
        super().__init__(func=func)

    def resolve(self, context):
        return self.func(**context)


class Attr(FrozenSlotted, Resolver):
    __slots__ = ("obj", "name")
    obj: Resolver
    name: str

    def __init__(self, obj, name):
        super().__init__(obj=resolver(obj), name=resolver(name))

    def __repr__(self):
        if isinstance(self.name, Just):
            return f"{self.obj!r}.{self.name.value}"
        else:
            return f"Attr({self.obj!r}, {self.name!r})"

    def resolve(self, context):
        obj = self.obj.resolve(context)
        name = self.name.resolve(context)
        return getattr(obj, name)


class Item(FrozenSlotted, Resolver):
    __slots__ = ("obj", "name")
    obj: Resolver
    name: str

    def __init__(self, obj, name):
        super().__init__(obj=resolver(obj), name=resolver(name))

    def __repr__(self):
        if isinstance(self.name, Just):
            return f"{self.obj!r}[{self.name.value!r}]"
        else:
            return f"Item({self.obj!r}, {self.name!r})"

    def resolve(self, context):
        obj = self.obj.resolve(context)
        name = self.name.resolve(context)
        return obj[name]


class Call(FrozenSlotted, Resolver):
    """Pattern that calls a function with the given arguments.

    Both positional and keyword arguments are coerced into patterns.

    Parameters
    ----------
    func
        The function to call.
    args
        The positional argument patterns.
    kwargs
        The keyword argument patterns.

    """

    __slots__ = ("func", "args", "kwargs")
    func: Resolver
    args: tuple[Resolver, ...]
    kwargs: dict[str, Resolver]

    def __init__(self, func, *args, **kwargs):
        if isinstance(func, Deferred):
            func = func._resolver
        elif isinstance(func, Resolver):
            pass
        elif callable(func):
            func = Just(func)
        else:
            raise TypeError(f"Invalid callable {func!r}")
        args = tuple(map(resolver, args))
        kwargs = FrozenDict({k: resolver(v) for k, v in kwargs.items()})
        super().__init__(func=func, args=args, kwargs=kwargs)

    def resolve(self, context):
        func = self.func.resolve(context)
        args = tuple(arg.resolve(context) for arg in self.args)
        kwargs = {k: v.resolve(context) for k, v in self.kwargs.items()}
        return func(*args, **kwargs)

    def __repr__(self):
        func = repr(self.func)
        args = ", ".join(map(repr, self.args))
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
        if args and kwargs:
            return f"{func}({args}, {kwargs})"
        elif args:
            return f"{func}({args})"
        elif kwargs:
            return f"{func}({kwargs})"
        else:
            return f"{func}()"


_operator_symbols = {
    operator.add: "+",
    operator.sub: "-",
    operator.mul: "*",
    operator.truediv: "/",
    operator.floordiv: "//",
    operator.pow: "**",
    operator.mod: "%",
    operator.eq: "==",
    operator.ne: "!=",
    operator.lt: "<",
    operator.le: "<=",
    operator.gt: ">",
    operator.ge: ">=",
    operator.and_: "&",
    operator.or_: "|",
    operator.xor: "^",
    operator.rshift: ">>",
    operator.lshift: "<<",
    operator.inv: "~",
    operator.neg: "-",
    operator.invert: "~",
}


class UnaryOperator(FrozenSlotted, Resolver):
    __slots__ = ("func", "arg")
    func: Callable
    arg: Resolver

    def __init__(self, func, arg):
        assert func in _operator_symbols
        super().__init__(func=func, arg=resolver(arg))

    def __repr__(self):
        symbol = _operator_symbols[self.func]
        return f"{symbol}{self.arg!r}"

    def resolve(self, context):
        arg = self.arg.resolve(context)
        return self.func(arg)


class BinaryOperator(FrozenSlotted, Resolver):
    __slots__ = ("func", "left", "right")
    func: Callable
    left: Resolver
    right: Resolver

    def __init__(self, func, left, right):
        assert func in _operator_symbols
        super().__init__(func=func, left=resolver(left), right=resolver(right))

    def __repr__(self):
        symbol = _operator_symbols[self.func]
        return f"({self.left!r} {symbol} {self.right!r})"

    def resolve(self, context):
        left = self.left.resolve(context)
        right = self.right.resolve(context)
        return self.func(left, right)


class Mapping(FrozenSlotted, Resolver):
    __slots__ = ("typ", "values")

    def __init__(self, values):
        typ = type(values)
        values = FrozenDict({k: resolver(v) for k, v in values.items()})
        super().__init__(typ=typ, values=values)

    def __repr__(self):
        items = ", ".join(f"{k!r}: {v!r}" for k, v in self.values.items())
        if self.typ is dict:
            return f"{{{items}}}"
        else:
            return f"{self.typ.__name__}({{{items}}})"

    def resolve(self, context):
        items = {k: v.resolve(context) for k, v in self.values.items()}
        return self.typ(items)


class Sequence(FrozenSlotted, Resolver):
    __slots__ = ("typ", "values")
    typ: type

    def __init__(self, values):
        typ = type(values)
        values = tuple(map(resolver, values))
        super().__init__(typ=typ, values=values)

    def __repr__(self):
        elems = ", ".join(map(repr, self.values))
        if self.typ is tuple:
            return f"({elems})"
        elif self.typ is list:
            return f"[{elems}]"
        else:
            return f"{self.typ.__name__}({elems})"

    def resolve(self, context):
        return self.typ(v.resolve(context) for v in self.values)


def resolver(obj):
    if isinstance(obj, Deferred):
        return obj._resolver
    elif isinstance(obj, Resolver):
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        # allow nesting deferred patterns in dicts
        return Mapping(obj)
    elif isinstance(obj, collections.abc.Sequence):
        # allow nesting deferred patterns in tuples/lists
        if isinstance(obj, (str, bytes)):
            return Just(obj)
        else:
            return Sequence(obj)
    else:
        # the object is used as a constant value
        return Just(obj)


def deferred(obj):
    return Deferred(resolver(obj))


def var(name):
    return Deferred(Variable(name))


def const(value):
    return Deferred(Just(value))


def _contains_deferred(obj: Any) -> bool:
    if isinstance(obj, (Resolver, Deferred)):
        return True
    elif (typ := type(obj)) in (tuple, list, set):
        return any(_contains_deferred(o) for o in obj)
    elif typ is dict:
        return any(_contains_deferred(o) for o in obj.values())
    return False


F = TypeVar("F", bound=Callable)


@overload
def deferrable(*, repr: str | None = None) -> Callable[[F], F]: ...


@overload
def deferrable(func: F) -> F: ...


def deferrable(func=None, *, repr=None):
    """Wrap a top-level expr function to support deferred arguments.

    When a deferrable function is called, the args & kwargs are traversed to
    look for `Deferred` values (through builtin collections like
    `list`/`tuple`/`set`/`dict`). If any `Deferred` arguments are found, then
    the result is also `Deferred`. Otherwise the function is called directly.

    Parameters
    ----------
    func
        A callable to make deferrable
    repr
        An optional fixed string to use when repr-ing the deferred expression,
        instead of the usual. This is useful for complex deferred expressions
        where the arguments don't necessarily make sense to be user facing
        in the repr.

    """

    def wrapper(func):
        # Parse the signature of func so we can validate deferred calls eagerly,
        # erroring for invalid/missing arguments at call time not resolve time.
        sig = inspect.signature(func)

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if _contains_deferred((args, kwargs)):
                # Try to bind the arguments now, raising a nice error
                # immediately if the function was called incorrectly
                sig.bind(*args, **kwargs)
                builder = Call(func, *args, **kwargs)
                return Deferred(builder, repr=repr)
            return func(*args, **kwargs)

        return inner  # type: ignore

    return wrapper if func is None else wrapper(func)


# reserved variable name for the value being matched
_ = var("_")



import itertools
import os
import re
import shlex
from pathlib import Path

import toolz
from attr import (
    evolve,
    field,
    frozen,
)
from attr.validators import (
    instance_of,
    optional,
)


env_templates_dir = Path(__file__).parents[2].joinpath("env_templates")


compiled_env_var_substitution_re = re.compile(r"^(?:\${(.*)}$)|(?:\$(.*))$")
compiled_env_var_setting_re = re.compile(
    "(?:export )?([^=]+)=(.*)",
    flags=re.DOTALL,
)


def parse_env_file(env_file, compiled_re=compiled_env_var_setting_re):
    def gen_shlex_lines(path):
        def make_lexer(path):
            lex = shlex.shlex(Path(path).read_text(), posix=True)
            lex.whitespace_split = True
            return lex

        def get_before_token_after(lexer):
            return (lexer.lineno, lexer.get_token(), lexer.lineno)

        tokens = ()
        for before, token, after in map(
            get_before_token_after,
            itertools.repeat(make_lexer(path)),
        ):
            if token is None:
                if tokens:
                    # single line env file never triggers `before != after`
                    yield " ".join(tokens)
                break
            tokens += (token,)
            if before != after:
                yield " ".join(tokens)
                tokens = ()

    matches = map(
        compiled_re.match,
        gen_shlex_lines(env_file),
    )
    dct = dict(match.groups() for match in filter(None, matches))
    return dct


@frozen
class EnvConfigable:
    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError as err:
            raise KeyError(key) from err

    def __contains__(self, key):
        return hasattr(self, key)

    def get(self, key, default=None):
        return getattr(self, key, default)

    clone = evolve

    def maybe_substitute_env_var(self, obj):
        return maybe_substitute_env_var(obj, self)

    @property
    def varnames(self):
        return tuple(self.get_defaults())

    @classmethod
    def get_env_overrides(cls, mapper=str.upper):
        dct = {
            name: value
            for (name, value) in (
                (attr.name, os.environ.get(mapper(attr.name)))
                for attr in cls.__attrs_attrs__
            )
            if value is not None
        }
        return dct

    @classmethod
    def get_defaults(cls):
        dct = {attr.name: attr.default for attr in cls.__attrs_attrs__}
        return dct

    @classmethod
    def from_env(cls, **kwargs):
        return cls(**(kwargs | cls.get_env_overrides()))

    @classmethod
    def subclass_from_kwargs(cls, *args, **kwargs):
        fields = {
            name: field(
                default=os.environ.get(name, ""), validator=optional(instance_of(str))
            )
            for name in args
        } | {
            name: field(
                default=value,
                validator=optional(instance_of(str if value is None else type(value))),
            )
            for name, value in kwargs.items()
        }
        return frozen(
            type(
                "EnvConfig",
                (cls,),
                fields,
            )
        )

    @classmethod
    def subclass_from_env_file(cls, env_file):
        env_file = Path(env_file).resolve()
        return cls.subclass_from_kwargs(
            **(
                parse_env_file(env_file)
                | {
                    "env_file": env_file,
                }
            )
        )


@toolz.curry
def maybe_substitute_env_var(obj, ctx=os.environ):
    if isinstance(obj, str) and (match := compiled_env_var_substitution_re.match(obj)):
        # this will match on "$"/"${}" and then raise on env_value is None
        env_var = next(filter(None, match.groups()), None)
        return ctx[env_var]
    else:
        return obj


def filter_existing_env_vars(dct, ctx):
    f = toolz.excepts(ValueError, maybe_substitute_env_var(ctx=ctx))
    dct = {k: v for k, v in dct.items() if (v is None or f(v) is not None)}
    return dct


def maybe_substitute_env_vars(dct: dict, ctx=os.environ) -> dict:
    return toolz.valmap(
        maybe_substitute_env_var(ctx=ctx),
        dct,
    )



import pandas as pd
import pytest


@pytest.fixture(scope="session")
def utf8_data():
    return pd.DataFrame(
        data=[("A", 1), ("B", 2), ("A", 2), ("A", 4), ("C", 1), ("A", 1)],
        columns=["name", "age"],
    )



"""User-defined functions (UDFs) implementation."""

from __future__ import annotations

import abc
import collections
import enum
import functools
import inspect
import itertools
import typing
from typing import TYPE_CHECKING, Any, Optional, TypeVar, overload

from public import public

import xorq.common.exceptions as exc
import xorq.vendor.ibis.expr.datashape as ds
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.annotations import Argument, attribute
from xorq.vendor.ibis.common.collections import FrozenDict
from xorq.vendor.ibis.common.deferred import deferrable


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, MutableMapping

    import xorq.vendor.ibis.expr.types as ir


EMPTY = inspect.Parameter.empty


_udf_name_cache: MutableMapping[type[ops.Node], Iterable[int]] = (
    collections.defaultdict(itertools.count)
)


def _make_udf_name(name: str) -> str:
    definition = next(_udf_name_cache[name])
    return f"{name}_{definition:d}"


@enum.unique
class InputType(enum.Enum):
    BUILTIN = enum.auto()
    PANDAS = enum.auto()
    PYARROW = enum.auto()
    PYTHON = enum.auto()


@public
class ScalarUDF(ops.Value):
    @attribute
    def shape(self):
        if not (args := getattr(self, "args")):  # noqa: B009
            # if a udf builtin takes no args then the shape check will fail
            # because there are no arguments to grab the shape of. In that case
            # default to a scalar shape
            return ds.scalar
        else:
            args = args if util.is_iterable(args) else [args]
            return rlz.highest_precedence_shape(args)


udf_meta_names = (
    "dtype",
    "__input_type__",
    "__func__",
    "__config__",
    "__udf_namespace__",
    "__module__",
    "__func_name__",
)


def restore_udf(name, bases, type_kwargs, ctor_kwargs):
    cls = type(name, bases, type_kwargs)
    return cls(**ctor_kwargs)


def reduce_udf(self, udf_meta_names=udf_meta_names):
    ctor_kwargs = dict(zip(self.argnames, self.args))
    fields = {
        name: Argument(pattern=rlz.ValueOf(value.dtype), typehint=value.dtype)
        for name, value in ctor_kwargs.items()
        if value is not None
    }
    meta = {k: getattr(self.__class__, k) for k in udf_meta_names}
    return restore_udf, (
        self.__class__.__name__,
        self.__class__.__bases__,
        fields | meta,
        ctor_kwargs,
    )


@public
class AggUDF(ops.Reduction):
    where: Optional[ops.Value[dt.Boolean]] = None

    __reduce__ = reduce_udf


def _wrap(
    wrapper,
    input_type: InputType,
    fn: Callable | None = None,
    **kwargs: Any,
) -> Callable:
    """Wrap a function `fn` with `wrapper`, allowing zero arguments when used as part of a decorator."""

    def wrap(fn):
        return functools.update_wrapper(
            deferrable(wrapper(input_type, fn, **kwargs)), fn
        )

    return wrap(fn) if fn is not None else wrap


S = TypeVar("S", bound=ops.Value)
B = TypeVar("B", bound=ops.Value)


class _UDF(abc.ABC):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def _base(self) -> type[B]:
        """Base class of the UDF."""

    @classmethod
    def _make_node(
        cls,
        fn: Callable,
        input_type: InputType,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple, Any] | None = None,
        **kwargs,
    ) -> type[S]:
        """Construct a scalar user-defined function that is built-in to the backend."""
        if "schema" in kwargs:
            raise exc.UnsupportedArgumentError(
                """schema` is not a valid argument.
                You can use the `catalog` and `database` keywords to specify a UDF location."""
            )

        if signature is None:
            annotations = typing.get_type_hints(fn)
            if (return_annotation := annotations.pop("return", None)) is None:
                raise exc.MissingReturnAnnotationError(fn)
            fields = {
                arg_name: Argument(
                    pattern=rlz.ValueOf(annotations.get(arg_name)),
                    default=param.default,
                    typehint=annotations.get(arg_name, Any),
                )
                for arg_name, param in inspect.signature(fn).parameters.items()
            }

        else:
            arg_types, return_annotation = signature
            arg_names = list(inspect.signature(fn).parameters)
            fields = {
                arg_name: Argument(pattern=rlz.ValueOf(typ), typehint=typ)
                for arg_name, typ in zip(arg_names, arg_types)
            }

        func_name = name if name is not None else fn.__name__

        fields.update(
            {
                "dtype": dt.dtype(return_annotation),
                "__input_type__": input_type,
                # must wrap `fn` in a `property` otherwise `fn` is assumed to be a
                # method
                "__func__": property(fget=lambda _, fn=fn: fn),
                "__config__": FrozenDict(kwargs),
                "__udf_namespace__": ops.Namespace(database=database, catalog=catalog),
                "__module__": fn.__module__,
                "__func_name__": func_name,
            }
        )

        return type(_make_udf_name(fn.__name__), (cls._base,), fields)

    @classmethod
    def _make_wrapper(
        cls, input_type: InputType, fn: Callable, **kwargs: Any
    ) -> Callable:
        node = cls._make_node(fn, input_type, **kwargs)

        @functools.wraps(fn)
        def construct(*args: Any, **kwargs: Any) -> ir.Value:
            return node(*args, **kwargs).to_expr()

        return construct


@public
class scalar(_UDF):
    """Scalar user-defined functions.

    ::: {.callout-note}
    ## The `scalar` class itself is **not** a public API, its methods are.
    :::
    """

    _base = ScalarUDF

    @overload
    @classmethod
    def builtin(cls, fn: Callable) -> Callable[..., ir.Value]: ...

    @overload
    @classmethod
    def builtin(
        cls,
        *,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple[Any, ...], Any] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable[..., ir.Value]]: ...

    @util.experimental
    @classmethod
    def builtin(
        cls,
        fn=None,
        *,
        name=None,
        database=None,
        catalog=None,
        signature=None,
        **kwargs,
    ):
        """Construct a scalar user-defined function that is built-in to the backend.

        Parameters
        ----------
        fn
            The function to wrap.
        name
            The name of the UDF in the backend if different from the function name.
        database
            The database in which the builtin function resides.
        catalog
            The catalog in which the builtin function resides.
        signature
            If present, a tuple of the form `((arg0type, arg1type, ...), returntype)`.
            For example, a function taking an int and a float and returning a
            string would be `((int, float), str)`. If not present, the signature
            will be derived from the type annotations of the wrapped function.

            For **builtin** UDFs, only the **return type** annotation is required.
            See [the user guide](/how-to/extending/builtin.qmd#input-types) for
            more information.
        kwargs
            Additional backend-specific configuration arguments for the UDF.

        Examples
        --------
        >>> import ibis
        >>> @ibis.udf.scalar.builtin
        ... def hamming(a: str, b: str) -> int:
        ...     '''Compute the Hamming distance between two strings.'''
        >>> expr = hamming("duck", "luck")
        >>> con = ibis.connect("duckdb://")
        >>> con.execute(expr)
        np.int64(1)

        """
        return _wrap(
            cls._make_wrapper,
            InputType.BUILTIN,
            fn,
            name=name,
            database=database,
            catalog=catalog,
            signature=signature,
            **kwargs,
        )

    @overload
    @classmethod
    def python(cls, fn: Callable) -> Callable[..., ir.Value]: ...

    @overload
    @classmethod
    def python(
        cls,
        *,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple[Any, ...], Any] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable[..., ir.Value]]: ...

    @util.experimental
    @classmethod
    def python(
        cls,
        fn=None,
        *,
        name=None,
        database=None,
        catalog=None,
        signature=None,
        **kwargs,
    ):
        """Construct a **non-vectorized** scalar user-defined function that accepts Python scalar values as inputs.

        ::: {.callout-warning collapse="true"}
        ## `python` UDFs are likely to be slow

        `python` UDFs are not vectorized: they are executed row by row with one
        Python function call per row

        This calling pattern tends to be **much** slower than
        [`pandas`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pandas)
        or
        [`pyarrow`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pyarrow)-based
        vectorized UDFs.
        :::

        Parameters
        ----------
        fn
            The function to wrap.
        name
            The name of the UDF in the backend if different from the function name.
        database
            The database in which to create the UDF.
        catalog
            The catalog in which to create the UDF.
        signature
            If present, a tuple of the form `((arg0type, arg1type, ...), returntype)`.
            For example, a function taking an int and a float and returning a
            string would be `((int, float), str)`. If not present, the signature
            will be derived from the type annotations of the wrapped function.
        kwargs
            Additional backend-specific configuration arguments for the UDF.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(dict(int_col=[1, 2, 3], str_col=["a", "b", "c"]))
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┓
        ┃ int_col ┃ str_col ┃
        ┡━━━━━━━━━╇━━━━━━━━━┩
        │ int64   │ string  │
        ├─────────┼─────────┤
        │       1 │ a       │
        │       2 │ b       │
        │       3 │ c       │
        └─────────┴─────────┘
        >>> @ibis.udf.scalar.python
        ... def str_magic(x: str) -> str:
        ...     return f"{x}_magic"
        >>> @ibis.udf.scalar.python
        ... def add_one_py(x: int) -> int:
        ...     return x + 1
        >>> str_magic(t.str_col)
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ str_magic_0(str_col) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ string               │
        ├──────────────────────┤
        │ a_magic              │
        │ b_magic              │
        │ c_magic              │
        └──────────────────────┘
        >>> add_one_py(t.int_col)
        ┏━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ add_one_py_0(int_col) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                 │
        ├───────────────────────┤
        │                     2 │
        │                     3 │
        │                     4 │
        └───────────────────────┘

        See Also
        --------
        - [`pandas`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pandas)
        - [`pyarrow`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pyarrow)

        """
        return _wrap(
            cls._make_wrapper,
            InputType.PYTHON,
            fn,
            name=name,
            database=database,
            catalog=catalog,
            signature=signature,
            **kwargs,
        )

    @overload
    @classmethod
    def pandas(cls, fn: Callable) -> Callable[..., ir.Value]: ...

    @overload
    @classmethod
    def pandas(
        cls,
        *,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple[Any, ...], Any] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable[..., ir.Value]]: ...

    @util.experimental
    @classmethod
    def pandas(
        cls,
        fn=None,
        *,
        name=None,
        database=None,
        catalog=None,
        signature=None,
        **kwargs,
    ):
        """Construct a **vectorized** scalar user-defined function that accepts pandas Series' as inputs.

        Parameters
        ----------
        fn
            The function to wrap.
        name
            The name of the UDF in the backend if different from the function name.
        database
            The database in which to create the UDF.
        catalog
            The catalog in which to create the UDF.
        signature
            If present, a tuple of the form `((arg0type, arg1type, ...), returntype)`.
            For example, a function taking an int and a float and returning a
            string would be `((int, float), str)`. If not present, the signature
            will be derived from the type annotations of the wrapped function.
        kwargs
            Additional backend-specific configuration arguments for the UDF.

        Examples
        --------
        ```python
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(dict(int_col=[1, 2, 3], str_col=["a", "b", "c"]))
        >>> t
        ┏━━━━━━━━━┳━━━━━━━━━┓
        ┃ int_col ┃ str_col ┃
        ┡━━━━━━━━━╇━━━━━━━━━┩
        │ int64   │ string  │
        ├─────────┼─────────┤
        │       1 │ a       │
        │       2 │ b       │
        │       3 │ c       │
        └─────────┴─────────┘
        >>> @ibis.udf.scalar.pandas
        ... def str_cap(x: str) -> str:
        ...     # note usage of pandas `str` method
        ...     return x.str.capitalize()
        >>> str_cap(t.str_col)  # doctest: +SKIP
        ┏━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ string_cap_0(str_col) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━┩
        │ string                │
        ├───────────────────────┤
        │ A                     │
        │ B                     │
        │ C                     │
        └───────────────────────┘
        ```

        See Also
        --------
        - [`python`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.python)
        - [`pyarrow`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pyarrow)

        """
        return _wrap(
            cls._make_wrapper,
            InputType.PANDAS,
            fn,
            name=name,
            database=database,
            catalog=catalog,
            signature=signature,
            **kwargs,
        )

    @overload
    @classmethod
    def pyarrow(cls, fn: Callable) -> Callable[..., ir.Value]: ...

    @overload
    @classmethod
    def pyarrow(
        cls,
        *,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple[Any, ...], Any] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable[..., ir.Value]]: ...

    @util.experimental
    @classmethod
    def pyarrow(
        cls,
        fn=None,
        *,
        name=None,
        database=None,
        catalog=None,
        signature=None,
        **kwargs,
    ):
        """Construct a **vectorized** scalar user-defined function that accepts PyArrow Arrays as input.

        Parameters
        ----------
        fn
            The function to wrap.
        name
            The name of the UDF in the backend if different from the function name.
        database
            The database in which to create the UDF.
        catalog
            The catalog in which to create the UDF.
        signature
            If present, a tuple of the form `((arg0type, arg1type, ...), returntype)`.
            For example, a function taking an int and a float and returning a
            string would be `((int, float), str)`. If not present, the signature
            will be derived from the type annotations of the wrapped function.
        kwargs
            Additional backend-specific configuration arguments for the UDF.

        Examples
        --------
        >>> import ibis
        >>> import pyarrow.compute as pc
        >>> from datetime import date
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(
        ...     dict(start_col=[date(2024, 4, 29)], end_col=[date(2025, 4, 29)]),
        ... )
        >>> @ibis.udf.scalar.pyarrow
        ... def weeks_between(start: date, end: date) -> int:
        ...     return pc.weeks_between(start, end)
        >>> weeks_between(t.start_col, t.end_col)
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ weeks_between_0(start_col, end_col) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                               │
        ├─────────────────────────────────────┤
        │                                  52 │
        └─────────────────────────────────────┘

        See Also
        --------
        - [`python`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.python)
        - [`pandas`](/reference/scalar-udfs.qmd#ibis.expr.operations.udf.scalar.pandas)

        """
        return _wrap(
            cls._make_wrapper,
            InputType.PYARROW,
            fn,
            name=name,
            database=database,
            catalog=catalog,
            signature=signature,
            **kwargs,
        )


@public
class agg(_UDF):
    """Aggregate user-defined functions.

    ::: {.callout-note}
    ## The `agg` class itself is **not** a public API, its methods are.
    :::
    """

    __slots__ = ()

    _base = AggUDF

    @overload
    @classmethod
    def builtin(cls, fn: Callable) -> Callable[..., ir.Value]: ...

    @overload
    @classmethod
    def builtin(
        cls,
        *,
        name: str | None = None,
        database: str | None = None,
        catalog: str | None = None,
        signature: tuple[tuple[Any, ...], Any] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable], Callable[..., ir.Value]]: ...

    @util.experimental
    @classmethod
    def builtin(
        cls,
        fn=None,
        *,
        name=None,
        database=None,
        catalog=None,
        signature=None,
        **kwargs,
    ):
        """Construct an aggregate user-defined function that is built-in to the backend.

        Parameters
        ----------
        fn
            The function to wrap.
        name
            The name of the UDF in the backend if different from the function name.
        database
            The database in which the builtin function resides.
        catalog
            The catalog in which the builtin function resides.
        signature
            If present, a tuple of the form `((arg0type, arg1type, ...), returntype)`.
            For example, a function taking an int and a float and returning a
            string would be `((int, float), str)`. If not present, the signature
            will be derived from the type annotations of the wrapped function.
        kwargs
            Additional backend-specific configuration arguments for the UDF.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> @ibis.udf.agg.builtin
        ... def favg(a: float) -> float:
        ...     '''Compute the average of a column using Kahan summation.'''
        >>> t = ibis.examples.penguins.fetch()
        >>> favg(t.bill_length_mm)
        ┌──────────┐
        │ 43.92193 │
        └──────────┘
        """
        return _wrap(
            cls._make_wrapper,
            InputType.BUILTIN,
            fn,
            name=name,
            database=database,
            catalog=catalog,
            signature=signature,
            **kwargs,
        )



import pytest
import toolz

import xorq.api as xo
from xorq.backends.snowflake import SnowflakeAuthenticator
from xorq.backends.snowflake.tests.conftest import (
    inside_temp_schema,
)
from xorq.vendor.ibis.util import gen_name


SU = pytest.importorskip("xorq.common.utils.snowflake_utils")


(database, schema) = ("SNOWFLAKE_SAMPLE_DATA", "TPCH_SF1")
make_connection = toolz.curry(
    SU.make_connection,
    authenticator=SnowflakeAuthenticator.keypair,
    database=database,
    schema=schema,
)


@pytest.mark.snowflake
def test_setup_session():
    con = make_connection(
        create_object_udfs=False,
    )
    dct = (
        con.raw_sql("SELECT CURRENT_WAREHOUSE(), CURRENT_DATABASE(), CURRENT_SCHEMA();")
        .fetch_pandas_all()
        .iloc[0]
        .to_dict()
    )
    expected = (database, schema)
    assert (con.current_catalog, con.current_database) == expected
    assert (con.con.database, con.con.schema) == expected
    assert dct == {
        "CURRENT_WAREHOUSE()": "COMPUTE_WH",
        "CURRENT_DATABASE()": database,
        "CURRENT_SCHEMA()": schema,
    }

    con = make_connection(
        create_object_udfs=True,
    )
    dct = (
        con.raw_sql("SELECT CURRENT_WAREHOUSE(), CURRENT_DATABASE(), CURRENT_SCHEMA();")
        .fetch_pandas_all()
        .iloc[0]
        .to_dict()
    )
    assert con.current_catalog == database
    assert con.current_database == schema
    assert con.con.database == database
    assert con.con.schema == schema
    assert dct == {
        "CURRENT_WAREHOUSE()": "COMPUTE_WH",
        "CURRENT_DATABASE()": database,
        "CURRENT_SCHEMA()": schema,
    }


@pytest.mark.snowflake
def test_table_namespace():
    con = make_connection()
    table = con.table("CUSTOMER")
    namespace = table.op().namespace
    assert namespace.catalog is not None
    assert namespace.database is not None


@pytest.mark.snowflake
def test_create_table_from_expr_failure(sf_con, temp_catalog, temp_db, csv_dir):
    name = "batting"
    t = xo.deferred_read_csv(csv_dir.joinpath(f"{name}.csv"), table_name=name)
    name = gen_name(t.get_name())
    with pytest.raises(ValueError, match="expr backend must be .*, is"):
        sf_con.create_table(name, t, database=f"{temp_catalog}.{temp_db}")


@pytest.mark.snowflake
def test_create_table_from_expr_success(sf_con, temp_catalog, temp_db, csv_dir):
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        name = "batting"
        t = xo.deferred_read_csv(
            csv_dir.joinpath(f"{name}.csv"), table_name=name
        ).into_backend(sf_con)
        name = gen_name(t.get_name())
        sf_con.create_table(name, t)
        assert name in sf_con.list_tables()


@pytest.mark.snowflake
def test_create_table_from_expr_other(
    sf_con, temp_catalog, temp_db, temp_catalog2, temp_db2, csv_dir
):
    with inside_temp_schema(sf_con, temp_catalog, temp_db):
        name = "batting"
        t = xo.deferred_read_csv(
            csv_dir.joinpath(f"{name}.csv"), table_name=name
        ).into_backend(sf_con)
        name = gen_name(t.get_name())
        assert not sf_con.list_tables()
        sf_con.create_table(name, t, database=f"{temp_catalog2}.{temp_db2}")
        assert name in sf_con.list_tables(database=(temp_catalog2, temp_db2))
